import time
import ctypes
import multiprocessing as mp
import itertools

import logging

from .util import shmem_as_ndarray, NP_TO_CTYPES
from .dsp_circuit import DSPCircuit
from .shared_ring_buffer import (ReadableSharedRingBuffer,
                                 WriteableSharedRingBuffer)

log = logging.getLogger(__name__)


def monitor(circuit_info, poll_period, pipe):
    '''
    Monitor loop that will be run in the subprocess.
    '''
    global RUN
    RUN = True
    log.debug("STARTING")

    ###########################################################################
    # Internal (private) function definitions
    ###########################################################################
    def process_request(pipe, circuits, timeout):
        # Check to see if data is available.  Keep checking until the timeout
        # period is over.
        if pipe.poll(timeout):
            id, device, request, args = pipe.recv()
            log.debug("Recieved request %s with args %r", request, args)

            # TERMINATE is a special request to end the process and shutdown.
            # We communicate this to the loop via a global variable, RUN.
            if request == 'TERMINATE':
                global RUN
                RUN = False
            else:
                # Process the request
                circuit = circuits[device]
                response = getattr(circuit, request)(*args)
                # Return the result of the request (along with the id)
                pipe.send((id, response))
                return False

    def read_buffer(hw_buffer, sw_buffer):
        if hw_buffer.pending():
            data = hw_buffer.read()
            # Timeout should be set to 0 (otherwise we are defeating the point
            # of using multiprocessing).
            sw_buffer.write(data, timeout=0)

    def write_buffer(hw_buffer, sw_buffer):
        if sw_buffer.should_set():
            data = sw_buffer.read()
            hw_buffer.set(data)
            sw_buffer._ioffset.value = -1
            log.debug("Request to set buffer %s with %d samples",
                      hw_buffer, len(data))
            sw_buffer.notify()
        elif hw_buffer.available() and sw_buffer.pending():
            log.debug("Writing data to buffer")
            available = hw_buffer.available()
            pending = sw_buffer.pending()
            samples = min((available, pending))
            data = sw_buffer.read(samples)
            hw_buffer.write(data[0])
            if not sw_buffer.pending():
                sw_buffer.notify()

    ###########################################################################
    # Initialization code
    ###########################################################################
    # Load the circuit first so we can use them to help initialize the buffers.
    # Once the circuit is loaded, initialize the bufers.  We do not currently
    # support read/write buffers (this would require locks, etc., and be
    # complicated to maintain and support).
    circuits = {}
    read_buffers, write_buffers = [], []
    for c_info, buffer_info in circuit_info.items():
        circuit_name, device_name = c_info
        circuit = DSPCircuit(circuit_name, device_name)
        circuits[device_name] = circuit

        for info in buffer_info:
            buffer_name = info['buffer_name']
            mode = info['mode']
            args = info['args']
            kwargs = info['kwargs']
            shmem = info['shmem']
            iwrite = info['iwrite']
            iread = info['iread']
            ioffset = info['ioffset']
            condition = info['condition']

            hw_buffer = circuit.get_buffer(buffer_name, mode, *args, **kwargs)
            cache = shmem_as_ndarray(shmem).reshape((hw_buffer.channels, -1))

            # The mode refers to the hardware buffer itself.  If we want to
            # read from the hardware buffer, then we need to write the data to
            # the shared memory via a WriteableSharedRingBuffer.  The other
            # process will be viewing the same memory space via a
            # ReadableSharedRingBuffer.
            args = cache, iwrite, iread, ioffset, condition, circuit
            if mode == 'r':
                sw_buffer = WriteableSharedRingBuffer(*args)
                read_buffers.append((hw_buffer, sw_buffer))
            elif mode == 'w':
                sw_buffer = ReadableSharedRingBuffer(*args)
                write_buffers.append((hw_buffer, sw_buffer))

        # Ok, now that we've initialized all the buffers for this circuit, it's
        # safe to start it.  Very important!  If you are running code that
        # requires multiple devices, then you need to be sure that the circuits
        # are designed to handle being started asynchronously (e.g. use the
        # zBUS trigger).
        circuit.start()

    # Notify the parent that the process has started
    pipe.send((None, 'STARTED'))

    ###########################################################################
    # Actual event loop
    ###########################################################################
    while RUN:
        log.debug("RUNNING")
        start = time.time()
        for hw_buffer, sw_buffer in read_buffers:
            # Since downloading buffer data can be a bit slow, let's check in
            # between downloads to see if we need to process any other requests
            # (i.e. set_tag and get_tag).
            process_request(pipe, circuits, 0)
            read_buffer(hw_buffer, sw_buffer)

        for hw_buffer, sw_buffer in write_buffers:
            process_request(pipe, circuits, 0)
            write_buffer(hw_buffer, sw_buffer)

        # Now that we're done polling the buffers, check to see how much time
        # has elapsed.  If all tasks were completed before the buffer poll
        # period has expired, shift task to checking request queue.
        while True:
            elapsed = time.time()-start
            if elapsed >= poll_period:
                break
            process_request(pipe, circuits, poll_period-elapsed)

    # When a terminate request has been recieved via the piplein, RUN is set to
    # False and we loop through all the circuits, stopping each one.
    for circuit in circuits.values():
        circuit.stop()

    # Notify the parent process that we have successfully terminated
    pipe.send((None, 'OK'))

    # Finally, the process exits.


class DSPProcess(mp.Process):

    def __init__(self, poll_period=0.1, cache_duration=30):
        self.cache_duration = cache_duration
        self.poll_period = poll_period

        # Shared information
        self._circuit_info = {}
        self._circuits = {}

        self._parent_pipe, self._child_pipe = mp.Pipe()
        self._request_id = itertools.count()

        import atexit
        atexit.register(self.terminate)
        super(DSPProcess, self).__init__(target=monitor)

    def load_circuit(self, circuit_name, device_name, device_id=1):
        '''
        Loads circuit and prepares shared memory for interprocess communication

        Parameters
        ----------
        circuit_name : str
            Path to circuit to load
        device_name : str
            Name of TDT System3 device to load circuit to
        device_id : number
            ID of device

        Returns
        -------
        circuit : instance of DSPCircuit
            The circuit.
        '''
        self._circuit_info[(circuit_name, device_name)] = []
        # We need to store a reference to the circuit here so we can properly
        # initialize any buffers we need
        circuit = DSPCircuit(circuit_name, device_name, device_id=device_id)
        self._circuits[device_name] = circuit
        shared_circuit = SharedCircuit(self, circuit_name, device_name)
        shared_circuit.fs = circuit.fs
        return shared_circuit

    def get_buffer(self, device_name, buffer_name, mode, *args, **kwargs):
        if self.is_alive():
            raise SystemError("Cannot allocate memory for buffer cache once "
                              "process has started")

        # Compute how much shared memory we need to allocate for the cache.  We
        # load the buffer in the parent process so we can inspect the circuit
        # and determine whether there are any pre-defined tags.
        circuit = self._circuits[device_name]

        buffer = circuit.get_buffer(buffer_name, mode, *args, **kwargs)
        cache_size = int(buffer.fs*self.cache_duration)*buffer.channels
        mem_type = NP_TO_CTYPES[buffer.dest_type]

        # Allocate the cache memory and create shared values used for
        # inter-process communication
        shmem = mp.RawArray(mem_type, cache_size)
        iwrite = mp.Value(ctypes.c_uint)
        iread = mp.Value(ctypes.c_uint)
        ioffset = mp.Value(ctypes.c_int)

        # Reentrant lock
        condition = mp.Condition(mp.RLock())

        # Save the information we need for reinitializing the buffer once the
        # process launches
        info = dict(
            circuit_name=circuit.name,
            device=circuit.device_name, buffer_name=buffer_name, mode=mode,
            args=args, kwargs=kwargs, shmem=shmem, iwrite=iwrite,
            iread=iread, ioffset=ioffset, condition=condition)
        key = (circuit.name, circuit.device_name)
        self._circuit_info[key].append(info)

        # Initialize the shared memory space for storing data acquired from the
        # DSP hardware
        cache = shmem_as_ndarray(shmem).reshape((buffer.channels, -1))
        args = cache, iwrite, iread, ioffset, condition, circuit
        if mode == 'r':
            sh_buffer = ReadableSharedRingBuffer(*args)
        elif mode == 'w':
            sh_buffer = WriteableSharedRingBuffer(*args)

        # Copy the attributes over from the actual buffer to the shared buffer
        # so we have access to the metadata we need
        attrs = ['fs', 'channels', 'data_tag']
        for k, v in buffer.attributes(attrs).items():
            setattr(sh_buffer, k, v)
        return sh_buffer

    def start(self):
        super(DSPProcess, self).start()
        # Wait until we have recieved a message that the process has
        # successfully launched.  If we do not recive a message after 10
        # seconds or recieve the wrong response, raise an error.
        if not self._parent_pipe.poll(10):
            raise SystemError("Unable to launch process")
        else:
            id, response = self._parent_pipe.recv()
            if response != 'STARTED':
                raise SystemError("Unable to launch process")

    def run(self):
        monitor(self._circuit_info, self.poll_period, self._child_pipe)

    def _get_response(self, device_name, request, args=None, timeout=None):
        '''
        Passes along request to the other process and waits for a response.
        '''
        # Generate a unique ID that ensures that the other process is
        # responding to the correct item
        if args is None:
            args = ()
        id = self._request_id.next()
        # Send the request to the process via a queue
        log.debug('Requesting %s:%s with arguments %r', device_name, request,
                  args)
        self._parent_pipe.send((id, device_name, request, args))
        # Wait till the response is available.  If no response is recieved
        # after one second, raise an exception.
        if not self._parent_pipe.poll(timeout):
            raise IOError("No response recieved for request %s(%r)"
                          % (request, args))
        # Get the response
        id, response = self._parent_pipe.recv()
        # Make sure it's the right response (this is probably an unecessary
        # check since we are blocking)
        if id != id:  # XXX This does nothing...?
            raise IOError("Wrong response returned")
        return response

    def stop(self):
        self._parent_pipe.send((None, None, 'TERMINATE', None))
        id, response = self._parent_pipe.recv()


def partial(f, device_name, action, timeout):
    def wrapper(*args):
        return f(device_name, action, args, timeout)
    return wrapper


class SharedCircuit(object):
    '''
    This is a duck-typed class meant to support the methods available on
    DSPCircuit.  It does not subclass DSPCircuit, but supports all the key
    methods you need (e.g. get_tag, set_tag, get_buffer).

    The only key difference is that you must ensure that shared memory for the
    process is allocated before you start it.  This means you must call
    `get_buffer` prior to calling `start`.  This is, in general, good practice
    anyway.
    '''

    # List of methods whose attributes can be sent directly to the DSP process
    # (pretty much all of them at the moment).  Note that set_coefficients may
    # not work for large coefficient tables, but we'll find out (the hard way)!
    PIPELINE_METHODS = ['stop', 'get_tag', 'set_tag', 'cget_tag', 'cset_tag',
                        'trigger', 'set_coefficients']

    def __init__(self, process, circuit_name, device_name):
        self.process = process
        self.device_name = device_name
        self.name = circuit_name

    def get_buffer(self, name, *args, **kwargs):
        return self.process.get_buffer(self.device_name, name, *args, **kwargs)

    def __getattr__(self, name):
        '''
        This is just wraps up the method call into an object that can be sent
        via the queue to the other process.
        '''
        if name in self.PIPELINE_METHODS:
            return partial(self.process._get_response, self.device_name,
                           name, 5)



# -*- coding: utf-8 -*-
# Created by makepy.py version 0.5.00
# By python version 2.6.6 |EPD 6.3-2 (32-bit)| (r266:84292, Sep 20 2010, 11:26:16) [MSC v.1500 32 bit (Intel)]
# From type library 'TTankX.ocx'
# On Thu Dec 02 21:01:19 2010
"""TTankX ActiveX Control module"""
makepy_version = '0.5.00'
python_version = 0x20606f0

import win32com.client.CLSIDToClass, pythoncom, pywintypes
import win32com.client.util
from pywintypes import IID
from win32com.client import Dispatch

# The following 3 lines may need tweaking for the particular server
# Candidates are pythoncom.Missing, .Empty and .ArgNotFound
defaultNamedOptArg=pythoncom.Empty
defaultNamedNotOptArg=pythoncom.Empty
defaultUnnamedArg=pythoncom.Empty

CLSID = IID('{3EABA0EF-2FBA-41F8-A970-3F238A4BAB01}')
MajorVersion = 1
MinorVersion = 0
LibraryFlags = 10
LCID = 0x0

from win32com.client import DispatchBaseClass
class _DTTankX(DispatchBaseClass):
	"""Dispatch interface for TTankX Control"""
	CLSID = IID('{BE6CAD3F-28F1-4EAC-B210-9CAA5CA8B5B8}')
	coclass_clsid = IID('{670490CE-57D2-4176-8E74-80C4C6A47D88}')

	def AboutBox(self):
		return self._oleobj_.InvokeTypes(-552, LCID, 1, (24, 0), (),)

	def AddClient(self, ClientName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(19, LCID, 1, (3, 0), ((8, 0),),ClientName
			)

	def AddServer(self, ServerName=defaultNamedNotOptArg, IPAddress=defaultNamedNotOptArg, Username=defaultNamedNotOptArg, Domain=defaultNamedNotOptArg
			, Password=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(48, LCID, 1, (3, 0), ((8, 0), (8, 0), (8, 0), (8, 0), (8, 0)),ServerName
			, IPAddress, Username, Domain, Password)

	def AddTank(self, TankName=defaultNamedNotOptArg, FilePath=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(20, LCID, 1, (3, 0), ((8, 0), (8, 0)),TankName
			, FilePath)

	def BuildEpocEv(self, Index=defaultNamedNotOptArg, TankCode=defaultNamedNotOptArg, TimeStamp=defaultNamedNotOptArg, Value=defaultNamedNotOptArg
			, BuddyEpoc=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(69, LCID, 1, (3, 0), ((3, 0), (8, 0), (5, 0), (5, 0), (8, 0)),Index
			, TankCode, TimeStamp, Value, BuddyEpoc)

	def BuildFilterDesc(self, TankCode=defaultNamedNotOptArg, TestCode=defaultNamedNotOptArg, V1=defaultNamedNotOptArg, V2=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(81, LCID, 1, (8, 0), ((3, 0), (3, 0), (5, 0), (5, 0)),TankCode
			, TestCode, V1, V2)

	def BuildScalar(self, Index=defaultNamedNotOptArg, TankCode=defaultNamedNotOptArg, SortChan=defaultNamedNotOptArg, TimeStamp=defaultNamedNotOptArg
			, Value=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(68, LCID, 1, (3, 0), ((3, 0), (8, 0), (3, 0), (5, 0), (5, 0)),Index
			, TankCode, SortChan, TimeStamp, Value)

	def BuildSnipEv(self, Index=defaultNamedNotOptArg, SortChan=defaultNamedNotOptArg, TimeStamp=defaultNamedNotOptArg, HasSort=defaultNamedNotOptArg
			, pData=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(71, LCID, 1, (3, 0), ((3, 0), (3, 0), (5, 0), (3, 0), (16388, 0)),Index
			, SortChan, TimeStamp, HasSort, pData)

	def BuildSnipEvV(self, Index=defaultNamedNotOptArg, SortChan=defaultNamedNotOptArg, TimeStamp=defaultNamedNotOptArg, HasSort=defaultNamedNotOptArg
			, vData=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(73, LCID, 1, (3, 0), ((3, 0), (3, 0), (5, 0), (3, 0), (16396, 0)),Index
			, SortChan, TimeStamp, HasSort, vData)

	def BuildStreamEv(self, Index=defaultNamedNotOptArg, SortChan=defaultNamedNotOptArg, TimeStamp=defaultNamedNotOptArg, pData=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(72, LCID, 1, (3, 0), ((3, 0), (3, 0), (5, 0), (16388, 0)),Index
			, SortChan, TimeStamp, pData)

	def BuildStreamEvV(self, Index=defaultNamedNotOptArg, SortChan=defaultNamedNotOptArg, TimeStamp=defaultNamedNotOptArg, vData=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(74, LCID, 1, (3, 0), ((3, 0), (3, 0), (5, 0), (12, 0)),Index
			, SortChan, TimeStamp, vData)

	def CheckTank(self, TankName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(21, LCID, 1, (3, 0), ((8, 0),),TankName
			)

	def ClearEpocIndexing(self):
		return self._oleobj_.InvokeTypes(85, LCID, 1, (24, 0), (),)

	def CloseTank(self):
		return self._oleobj_.InvokeTypes(22, LCID, 1, (24, 0), (),)

	def CodeToString(self, EvCode=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(23, LCID, 1, (8, 0), ((3, 0),),EvCode
			)

	def ConnectServer(self, ServerName=defaultNamedNotOptArg, ClientName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(18, LCID, 1, (3, 0), ((8, 0), (8, 0)),ServerName
			, ClientName)

	def CreateEpocIndexing(self):
		return self._oleobj_.InvokeTypes(84, LCID, 1, (3, 0), (),)

	def DFromToString(self, DFormCode=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(27, LCID, 1, (8, 0), ((3, 0),),DFormCode
			)

	def DeleteClient(self, ClientName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(24, LCID, 1, (24, 0), ((8, 0),),ClientName
			)

	def DeleteSortCode(self, sortName=defaultNamedNotOptArg, snipName=defaultNamedNotOptArg, idxChan=defaultNamedNotOptArg):
		"""method DeleteSortCode"""
		return self._oleobj_.InvokeTypes(125, LCID, 1, (3, 0), ((8, 0), (8, 0), (3, 0)),sortName
			, snipName, idxChan)

	def DisableTankDebug(self, TankName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(34, LCID, 1, (24, 0), ((8, 0),),TankName
			)

	def EnableTankDebug(self, TankName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(33, LCID, 1, (24, 0), ((8, 0),),TankName
			)

	def EvTypeToString(self, evTypeCode=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(26, LCID, 1, (8, 0), ((3, 0),),evTypeCode
			)

	def FancyTime(self, Time=defaultNamedNotOptArg, Format=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(44, LCID, 1, (8, 0), ((5, 0), (8, 0)),Time
			, Format)

	def FromTTD(self, TTD=defaultNamedNotOptArg, FieldCode=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(54, LCID, 1, (8, 0), ((8, 0), (8, 0)),TTD
			, FieldCode)

	def GetClientID(self, ClientName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(35, LCID, 1, (3, 0), ((8, 0),),ClientName
			)

	def GetCodeSpecs(self, EvCode=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(36, LCID, 1, (3, 0), ((3, 0),),EvCode
			)

	def GetCodeSpecsLazy(self, EvCode=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(111, LCID, 1, (3, 0), ((3, 0),),EvCode
			)

	def GetDebug(self, TankName=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(28, LCID, 1, (8, 0), ((8, 0),),TankName
			)

	def GetEnumServer(self, Index=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(51, LCID, 1, (8, 0), ((3, 0),),Index
			)

	def GetEnumTank(self, Index=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(47, LCID, 1, (8, 0), ((3, 0),),Index
			)

	def GetEpocCode(self, Index=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(58, LCID, 1, (8, 0), ((3, 0),),Index
			)

	def GetEpocs(self, pEPs=defaultNamedNotOptArg, TankCode=defaultNamedNotOptArg, T1=defaultNamedNotOptArg, T2=defaultNamedNotOptArg
			, MaxRet=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(60, LCID, 1, (3, 0), ((16389, 0), (3, 0), (5, 0), (5, 0), (3, 0)),pEPs
			, TankCode, T1, T2, MaxRet)

	def GetEpocsEx(self, pEPs=defaultNamedNotOptArg, TankCode=defaultNamedNotOptArg, MaxReturn=defaultNamedNotOptArg, Mode=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(94, LCID, 1, (3, 0), ((16389, 0), (3, 0), (3, 0), (3, 0)),pEPs
			, TankCode, MaxReturn, Mode)

	def GetEpocsExV(self, TankCode=defaultNamedNotOptArg, Mode=defaultNamedNotOptArg):
		return self._ApplyTypes_(95, 1, (12, 0), ((8, 0), (3, 0)), u'GetEpocsExV', None,TankCode
			, Mode)

	def GetEpocsV(self, TankCode=defaultNamedNotOptArg, T1=defaultNamedNotOptArg, T2=defaultNamedNotOptArg, MaxEpocs=defaultNamedNotOptArg):
		return self._ApplyTypes_(57, 1, (12, 0), ((8, 0), (5, 0), (5, 0), (3, 0)), u'GetEpocsV', None,TankCode
			, T1, T2, MaxEpocs)

	def GetError(self):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(29, LCID, 1, (8, 0), (),)

	def GetEvTsqIdx(self):
		"""method GetEvTsqIdx"""
		return self._ApplyTypes_(122, 1, (12, 0), (), u'GetEvTsqIdx', None,)

	def GetEventCodes(self, EvType=defaultNamedNotOptArg):
		return self._ApplyTypes_(38, 1, (12, 0), ((3, 0),), u'GetEventCodes', None,EvType
			)

	def GetFilterTolerance(self):
		return self._oleobj_.InvokeTypes(87, LCID, 1, (5, 0), (),)

	def GetGlobal(self, code=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(106, LCID, 1, (5, 0), ((3, 0),),code
			)

	def GetGlobalB(self, name=defaultNamedNotOptArg):
		"""method GetGlobalB"""
		return self._oleobj_.InvokeTypes(119, LCID, 1, (5, 0), ((8, 0),),name
			)

	def GetGlobalStringB(self, name=defaultNamedNotOptArg):
		"""method GetGlobalStringB"""
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(120, LCID, 1, (8, 0), ((8, 0),),name
			)

	def GetGlobalStringV(self, name=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(108, LCID, 1, (8, 0), ((31, 0),),name
			)

	def GetGlobalV(self, name=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(107, LCID, 1, (5, 0), ((31, 0),),name
			)

	def GetHotBlock(self):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(66, LCID, 1, (8, 0), (),)

	def GetNPer(self, DForm=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(59, LCID, 1, (3, 0), ((3, 0),),DForm
			)

	def GetServerItem(self, ServerName=defaultNamedNotOptArg, ItemCode=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(50, LCID, 1, (8, 0), ((8, 0), (8, 0)),ServerName
			, ItemCode)

	def GetSortChanMap(self, sortName=defaultNamedNotOptArg, snipName=defaultNamedNotOptArg):
		"""method GetSortChanMap"""
		return self._ApplyTypes_(126, 1, (12, 0), ((8, 0), (8, 0)), u'GetSortChanMap', None,sortName
			, snipName)

	def GetSortCondition(self, sortName=defaultNamedNotOptArg, snipName=defaultNamedNotOptArg, idxChan=defaultNamedNotOptArg):
		"""method GetSortCondition"""
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(124, LCID, 1, (8, 0), ((8, 0), (8, 0), (3, 0)),sortName
			, snipName, idxChan)

	def GetSortName(self, EventName=defaultNamedNotOptArg, idxSortID=defaultNamedNotOptArg):
		"""method GetSortName"""
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(113, LCID, 1, (8, 0), ((8, 0), (3, 0)),EventName
			, idxSortID)

	def GetStatus(self, StatCode=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(39, LCID, 1, (3, 0), ((3, 0),),StatCode
			)

	def GetTankItem(self, TankName=defaultNamedNotOptArg, ItemCode=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(52, LCID, 1, (8, 0), ((8, 0), (8, 0)),TankName
			, ItemCode)

	def GetValidTimeRanges(self, pRanges=defaultNamedNotOptArg, MaxReturn=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(99, LCID, 1, (3, 0), ((16389, 0), (3, 0)),pRanges
			, MaxReturn)

	def GetValidTimeRangesV(self):
		return self._ApplyTypes_(100, 1, (12, 0), (), u'GetValidTimeRangesV', None,)

	def IndexEvent(self, TankCode=defaultNamedNotOptArg, Channel=defaultNamedNotOptArg, SortCode=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(83, LCID, 1, (3, 0), ((3, 0), (3, 0), (3, 0)),TankCode
			, Channel, SortCode)

	def InitializeTank(self, TankName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(40, LCID, 1, (3, 0), ((8, 0),),TankName
			)

	def OpenTank(self, TankName=defaultNamedNotOptArg, AccessMode=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(30, LCID, 1, (3, 0), ((8, 0), (8, 0)),TankName
			, AccessMode)

	def ParseEv(self, RecIndex=defaultNamedNotOptArg, TimeStamp=defaultNamedNotOptArg, Channel=defaultNamedNotOptArg, SortCode=defaultNamedNotOptArg
			, Npts=defaultNamedNotOptArg, pData=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(75, LCID, 1, (3, 0), ((3, 0), (16389, 0), (16387, 0), (16387, 0), (16387, 0), (16388, 0)),RecIndex
			, TimeStamp, Channel, SortCode, Npts, pData
			)

	def ParseEvInfoV(self, RecIndex=defaultNamedNotOptArg, nRecs=defaultNamedNotOptArg, nItem=defaultNamedNotOptArg):
		return self._ApplyTypes_(77, 1, (12, 0), ((3, 0), (3, 0), (3, 0)), u'ParseEvInfoV', None,RecIndex
			, nRecs, nItem)

	def ParseEvV(self, RecIndex=defaultNamedNotOptArg, nRecs=defaultNamedNotOptArg):
		return self._ApplyTypes_(76, 1, (12, 0), ((3, 0), (3, 0)), u'ParseEvV', None,RecIndex
			, nRecs)

	def ParseFilterDesc(self, FiltDesc=defaultNamedNotOptArg, TankCode=defaultNamedNotOptArg, TestCode=defaultNamedNotOptArg, V1=defaultNamedNotOptArg
			, V2=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(80, LCID, 1, (3, 0), ((8, 0), (16387, 0), (16387, 0), (16389, 0), (16389, 0)),FiltDesc
			, TankCode, TestCode, V1, V2)

	def QryEpocAt(self, TankCode=defaultNamedNotOptArg, rTime=defaultNamedNotOptArg, ReqItem=defaultNamedNotOptArg, RetVal=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(64, LCID, 1, (3, 0), ((3, 0), (5, 0), (3, 0), (16389, 0)),TankCode
			, rTime, ReqItem, RetVal)

	def QryEpocAtV(self, TankCode=defaultNamedNotOptArg, rTime=defaultNamedNotOptArg, ReqItem=defaultNamedNotOptArg):
		return self._ApplyTypes_(65, 1, (12, 0), ((8, 0), (5, 0), (3, 0)), u'QryEpocAtV', None,TankCode
			, rTime, ReqItem)

	def QueryBlockName(self, BlockNumber=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(41, LCID, 1, (8, 0), ((3, 0),),BlockNumber
			)

	def ReadEvents(self, MaxRet=defaultNamedNotOptArg, TankCode=defaultNamedNotOptArg, Channel=defaultNamedNotOptArg, SortCode=defaultNamedNotOptArg
			, T1=defaultNamedNotOptArg, T2=defaultNamedNotOptArg, Options=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(62, LCID, 1, (3, 0), ((3, 0), (3, 0), (3, 0), (3, 0), (5, 0), (5, 0), (3, 0)),MaxRet
			, TankCode, Channel, SortCode, T1, T2
			, Options)

	def ReadEventsSimple(self, EventName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(96, LCID, 1, (3, 0), ((8, 0),),EventName
			)

	def ReadEventsV(self, MaxRet=defaultNamedNotOptArg, TankCode=defaultNamedNotOptArg, Channel=defaultNamedNotOptArg, SortCode=defaultNamedNotOptArg
			, T1=defaultNamedNotOptArg, T2=defaultNamedNotOptArg, Options=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(63, LCID, 1, (3, 0), ((3, 0), (8, 0), (3, 0), (3, 0), (5, 0), (5, 0), (8, 0)),MaxRet
			, TankCode, Channel, SortCode, T1, T2
			, Options)

	def ReadWaves(self, TankCode=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(97, LCID, 1, (3, 0), ((3, 0),),TankCode
			)

	def ReadWavesOnTimeRange(self, TankCode=defaultNamedNotOptArg, Channel=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(101, LCID, 1, (3, 0), ((3, 0), (3, 0)),TankCode
			, Channel)

	def ReadWavesOnTimeRangeB(self, TankCode=defaultNamedNotOptArg, Channel=defaultNamedNotOptArg):
		"""method ReadWavesOnTimeRangeB"""
		return self._ApplyTypes_(116, 1, (12, 0), ((8, 0), (3, 0)), u'ReadWavesOnTimeRangeB', None,TankCode
			, Channel)

	def ReadWavesOnTimeRangeV(self, TankCode=defaultNamedNotOptArg, Channel=defaultNamedNotOptArg):
		return self._ApplyTypes_(102, 1, (12, 0), ((31, 0), (3, 0)), u'ReadWavesOnTimeRangeV', None,TankCode
			, Channel)

	def ReadWavesV(self, TankCode=defaultNamedNotOptArg):
		return self._ApplyTypes_(98, 1, (12, 0), ((8, 0),), u'ReadWavesV', None,TankCode
			)

	def ReleaseServer(self):
		return self._oleobj_.InvokeTypes(42, LCID, 1, (24, 0), (),)

	def RemoveBlock(self, BlockName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(56, LCID, 1, (3, 0), ((8, 0),),BlockName
			)

	def RemoveEvents(self, BlockName=defaultNamedNotOptArg, RelTime1=defaultNamedNotOptArg, RelTime2=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(55, LCID, 1, (3, 0), ((8, 0), (5, 0), (5, 0)),BlockName
			, RelTime1, RelTime2)

	def RemoveServer(self, ServerName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(49, LCID, 1, (3, 0), ((8, 0),),ServerName
			)

	def RemoveTank(self, TankName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(25, LCID, 1, (3, 0), ((8, 0),),TankName
			)

	def ResetFilters(self):
		return self._oleobj_.InvokeTypes(82, LCID, 1, (24, 0), (),)

	def ResetGlobals(self):
		return self._oleobj_.InvokeTypes(110, LCID, 1, (24, 0), (),)

	def ResetTank(self, TankName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(43, LCID, 1, (3, 0), ((8, 0),),TankName
			)

	def SaveSortCodes(self, sortName=defaultNamedNotOptArg, snipName=defaultNamedNotOptArg, idxChan=defaultNamedNotOptArg, sortCondition=defaultNamedNotOptArg
			, sortCodeArray=defaultNamedNotOptArg):
		"""method SaveSortCodes"""
		return self._oleobj_.InvokeTypes(123, LCID, 1, (3, 0), ((8, 0), (8, 0), (3, 0), (8, 0), (12, 0)),sortName
			, snipName, idxChan, sortCondition, sortCodeArray)

	def SelectBlock(self, BlockName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(32, LCID, 1, (3, 0), ((8, 0),),BlockName
			)

	def SetBuildHead(self, TankCode=defaultNamedNotOptArg, DataForm=defaultNamedNotOptArg, SampFreq=defaultNamedNotOptArg, nNet=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(70, LCID, 1, (24, 0), ((8, 0), (3, 0), (5, 0), (3, 0)),TankCode
			, DataForm, SampFreq, nNet)

	def SetEpocTimeFilter(self, EpocCode=defaultNamedNotOptArg, Offset=defaultNamedNotOptArg, Dur=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(90, LCID, 1, (3, 0), ((3, 0), (5, 0), (5, 0)),EpocCode
			, Offset, Dur)

	def SetEpocTimeFilterB(self, EpocCode=defaultNamedNotOptArg, Offset=defaultNamedNotOptArg, Dur=defaultNamedNotOptArg):
		"""method SetEpocTimeFilterB"""
		return self._oleobj_.InvokeTypes(114, LCID, 1, (3, 0), ((8, 0), (5, 0), (5, 0)),EpocCode
			, Offset, Dur)

	def SetEpocTimeFilterV(self, EpocCode=defaultNamedNotOptArg, Offset=defaultNamedNotOptArg, Dur=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(91, LCID, 1, (3, 0), ((31, 0), (5, 0), (5, 0)),EpocCode
			, Offset, Dur)

	def SetFilter(self, TankCode=defaultNamedNotOptArg, TestCode=defaultNamedNotOptArg, V1=defaultNamedNotOptArg, V2=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(78, LCID, 1, (3, 0), ((3, 0), (3, 0), (5, 0), (5, 0)),TankCode
			, TestCode, V1, V2)

	def SetFilterArray(self, Dim=defaultNamedNotOptArg, ID=defaultNamedNotOptArg, Filter=defaultNamedNotOptArg, Exclusive=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(89, LCID, 1, (3, 0), ((3, 0), (3, 0), (8, 0), (3, 0)),Dim
			, ID, Filter, Exclusive)

	def SetFilterTolerance(self, loosenFactor=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(86, LCID, 1, (3, 0), ((5, 0),),loosenFactor
			)

	def SetFilterWithDesc(self, FiltDesc=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(79, LCID, 1, (3, 0), ((8, 0),),FiltDesc
			)

	def SetFilterWithDescEx(self, Filter=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(88, LCID, 1, (3, 0), ((8, 0),),Filter
			)

	def SetGlobal(self, code=defaultNamedNotOptArg, Value=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(103, LCID, 1, (3, 0), ((3, 0), (5, 0)),code
			, Value)

	def SetGlobalB(self, name=defaultNamedNotOptArg, Value=defaultNamedNotOptArg):
		"""method SetGlobalB"""
		return self._oleobj_.InvokeTypes(117, LCID, 1, (3, 0), ((8, 0), (5, 0)),name
			, Value)

	def SetGlobalStringB(self, name=defaultNamedNotOptArg, strvalue=defaultNamedNotOptArg):
		"""method SetGlobalStringB"""
		return self._oleobj_.InvokeTypes(118, LCID, 1, (3, 0), ((8, 0), (8, 0)),name
			, strvalue)

	def SetGlobalStringV(self, name=defaultNamedNotOptArg, strvalue=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(105, LCID, 1, (3, 0), ((31, 0), (31, 0)),name
			, strvalue)

	def SetGlobalV(self, name=defaultNamedNotOptArg, Value=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(104, LCID, 1, (3, 0), ((31, 0), (5, 0)),name
			, Value)

	def SetGlobals(self, str=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(109, LCID, 1, (3, 0), ((31, 0),),str
			)

	def SetGlobalsB(self, str=defaultNamedNotOptArg):
		"""method SetGlobalsB"""
		return self._oleobj_.InvokeTypes(121, LCID, 1, (3, 0), ((8, 0),),str
			)

	def SetRefEpoc(self, EpocCode=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(92, LCID, 1, (3, 0), ((3, 0),),EpocCode
			)

	def SetRefEpocB(self, EpocCode=defaultNamedNotOptArg):
		"""method SetRefEpocB"""
		return self._oleobj_.InvokeTypes(115, LCID, 1, (3, 0), ((8, 0),),EpocCode
			)

	def SetRefEpocV(self, EpocCode=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(93, LCID, 1, (3, 0), ((31, 0),),EpocCode
			)

	def SetRefTime(self, TimeOS=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(31, LCID, 1, (3, 0), ((5, 0),),TimeOS
			)

	def SetUseSortName(self, sortID=defaultNamedNotOptArg):
		"""method SetUseSortName"""
		return self._oleobj_.InvokeTypes(112, LCID, 1, (3, 0), ((8, 0),),sortID
			)

	def StartRecord(self, BlockName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(45, LCID, 1, (3, 0), ((8, 0),),BlockName
			)

	def StopRecord(self):
		return self._oleobj_.InvokeTypes(46, LCID, 1, (24, 0), (),)

	def StringToEvCode(self, EvCode=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(37, LCID, 1, (3, 0), ((8, 0),),EvCode
			)

	def SwitchClient(self, ClientName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(67, LCID, 1, (3, 0), ((8, 0),),ClientName
			)

	def ToTTD(self, ServName=defaultNamedNotOptArg, TankName=defaultNamedNotOptArg, BlockName=defaultNamedNotOptArg, EvName=defaultNamedNotOptArg
			, ChanList=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(53, LCID, 1, (8, 0), ((8, 0), (8, 0), (8, 0), (8, 0), (16387, 0)),ServName
			, TankName, BlockName, EvName, ChanList)

	def WriteEvents(self, NumEv=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(61, LCID, 1, (3, 0), ((3, 0),),NumEv
			)

	_prop_map_get_ = {
		"ClientID": (6, 2, (3, 0), (), "ClientID", None),
		"CurBlockMemo": (3, 2, (8, 0), (), "CurBlockMemo", None),
		"CurBlockName": (1, 2, (8, 0), (), "CurBlockName", None),
		"CurBlockNotes": (17, 2, (8, 0), (), "CurBlockNotes", None),
		"CurBlockOwner": (2, 2, (8, 0), (), "CurBlockOwner", None),
		"CurBlockStartTime": (4, 2, (5, 0), (), "CurBlockStartTime", None),
		"CurBlockStopTime": (5, 2, (5, 0), (), "CurBlockStopTime", None),
		"Domain": (15, 2, (8, 0), (), "Domain", None),
		"EvChannel": (13, 2, (3, 0), (), "EvChannel", None),
		"EvDForm": (10, 2, (3, 0), (), "EvDForm", None),
		"EvDataSize": (8, 2, (3, 0), (), "EvDataSize", None),
		"EvFirstTime": (9, 2, (5, 0), (), "EvFirstTime", None),
		"EvSampFreq": (11, 2, (4, 0), (), "EvSampFreq", None),
		"EvTimeStamp": (12, 2, (5, 0), (), "EvTimeStamp", None),
		"EvType": (7, 2, (3, 0), (), "EvType", None),
		"Password": (16, 2, (8, 0), (), "Password", None),
		"Username": (14, 2, (8, 0), (), "Username", None),
	}
	_prop_map_put_ = {
		"ClientID" : ((6, LCID, 4, 0),()),
		"CurBlockMemo" : ((3, LCID, 4, 0),()),
		"CurBlockName" : ((1, LCID, 4, 0),()),
		"CurBlockNotes" : ((17, LCID, 4, 0),()),
		"CurBlockOwner" : ((2, LCID, 4, 0),()),
		"CurBlockStartTime" : ((4, LCID, 4, 0),()),
		"CurBlockStopTime" : ((5, LCID, 4, 0),()),
		"Domain" : ((15, LCID, 4, 0),()),
		"EvChannel" : ((13, LCID, 4, 0),()),
		"EvDForm" : ((10, LCID, 4, 0),()),
		"EvDataSize" : ((8, LCID, 4, 0),()),
		"EvFirstTime" : ((9, LCID, 4, 0),()),
		"EvSampFreq" : ((11, LCID, 4, 0),()),
		"EvTimeStamp" : ((12, LCID, 4, 0),()),
		"EvType" : ((7, LCID, 4, 0),()),
		"Password" : ((16, LCID, 4, 0),()),
		"Username" : ((14, LCID, 4, 0),()),
	}

class _DTTankXEvents:
	"""Event interface for TTankX Control"""
	CLSID = CLSID_Sink = IID('{9C0BD59B-8842-47DA-8105-DC3E883D72D1}')
	coclass_clsid = IID('{670490CE-57D2-4176-8E74-80C4C6A47D88}')
	_public_methods_ = [] # For COM Server support
	_dispid_to_func_ = {
		}

	def __init__(self, oobj = None):
		if oobj is None:
			self._olecp = None
		else:
			import win32com.server.util
			from win32com.server.policy import EventHandlerPolicy
			cpc=oobj._oleobj_.QueryInterface(pythoncom.IID_IConnectionPointContainer)
			cp=cpc.FindConnectionPoint(self.CLSID_Sink)
			cookie=cp.Advise(win32com.server.util.wrap(self, usePolicy=EventHandlerPolicy))
			self._olecp,self._olecp_cookie = cp,cookie
	def __del__(self):
		try:
			self.close()
		except pythoncom.com_error:
			pass
	def close(self):
		if self._olecp is not None:
			cp,cookie,self._olecp,self._olecp_cookie = self._olecp,self._olecp_cookie,None,None
			cp.Unadvise(cookie)
	def _query_interface_(self, iid):
		import win32com.server.util
		if iid==self.CLSID_Sink: return win32com.server.util.wrap(self)

	# Event Handlers
	# If you create handlers, they should have the following prototypes:


from win32com.client import CoClassBaseClass
# This CoClass is known by the name 'TTANK.X'
class TTankX(CoClassBaseClass): # A CoClass
	# TTankX Control
	CLSID = IID('{670490CE-57D2-4176-8E74-80C4C6A47D88}')
	coclass_sources = [
		_DTTankXEvents,
	]
	default_source = _DTTankXEvents
	coclass_interfaces = [
		_DTTankX,
	]
	default_interface = _DTTankX

RecordMap = {
}

CLSIDToClassMap = {
	'{670490CE-57D2-4176-8E74-80C4C6A47D88}' : TTankX,
	'{9C0BD59B-8842-47DA-8105-DC3E883D72D1}' : _DTTankXEvents,
	'{BE6CAD3F-28F1-4EAC-B210-9CAA5CA8B5B8}' : _DTTankX,
}
CLSIDToPackageMap = {}
win32com.client.CLSIDToClass.RegisterCLSIDsFromDict( CLSIDToClassMap )
VTablesToPackageMap = {}
VTablesToClassMap = {
}


NamesToIIDMap = {
	'_DTTankX' : '{BE6CAD3F-28F1-4EAC-B210-9CAA5CA8B5B8}',
	'_DTTankXEvents' : '{9C0BD59B-8842-47DA-8105-DC3E883D72D1}',
}





'''
Functions for loading the zBUS, PA5 and RPcoX drivers and connecting to the
specified device.  In addition to loading the appropriate ActiveX driver, some
minimal configuration is done.

Network-aware proxies of the zBUS and RPcoX drivers have been written for
TDTPy. To connect to TDT hardware that is running on a remote computer, both
the :func:`connect_zbus` and :func:`connect_rpcox` functions take the address
of the server via a tuple (hostname, port)::

    connect_rpcox('RZ6', address=(tdt_server.cns.nyu.edu, 3333))

.. autofunction:: connect_zbus
.. autofunction:: connect_rpcox
.. autofunction:: connect_pa5

.. note::

    The network-aware proxy code should be considered alpha stage.  Although it
    appears to work in our tests, we have not deployed this in our data
    aqcuisition experiments.

'''
import os
import numpy as np
import ctypes

# Initialize
from .dsp_error import DSPError
from . import dsp_server

import logging
log = logging.getLogger(__name__)


def connect_pa5(interface='GB', device_id=1, address=None):
    '''
    Connect to the PA5
    '''
    import pythoncom
    import pywintypes
    from . import actxobjects
    debug_string = '%d via %s interface' % (device_id, interface)
    log.debug(debug_string)
    try:
        pythoncom.CoInitialize()
        if address is None:
            driver = actxobjects.PA5x()
        else:
            driver = dsp_server.PA5NET(address)
        if not driver.ConnectPA5(interface, device_id):
            raise DSPError("PA5", "Connection failed")
        log.debug("Connected to PA5")

        return driver
    except pywintypes.com_error:
        raise ImportError('ActiveX drivers from TDT not installed')


def connect_zbus(interface='GB', address=None):
    '''
    Connect to the zBUS interface and set the zBUS A and zBUS B triggers to low

    Parameters
    ----------
    interface : {'GB', 'USB'}
        Type of interface (depends on the card that you have from TDT). See the
        TDT ActiveX documentation for clarification on which interface you
        would be using if you are still unsure.
    address : {None, (hostname, port)}
        If None, loads the ActiveX drivers directly, otherwise connects to the
        remote server specified by the hostname, port tuple.
    '''
    import pythoncom
    import pywintypes
    from . import actxobjects
    try:
        pythoncom.CoInitialize()
        if address is not None:
            driver = dsp_server.zBUSNET(address)
        else:
            driver = actxobjects.ZBUSx()
        if not driver.ConnectZBUS(interface):
            raise DSPError("zBUS", "Connection failed")
        log.debug("Connected to zBUS")

        # zBUS trigger is set to high for record mode, so ensure that both
        # triggers are initialized to low.
        driver.zBusTrigA(0, 2, 10)
        driver.zBusTrigB(0, 2, 10)
        log.debug("Set zBusTrigA to low")
        log.debug("Set zBusTrigB to low")
        return driver
    except pywintypes.com_error:
        raise ImportError('ActiveX drivers from TDT not installed')


def connect_rpcox(name, interface='GB', device_id=1, address=None):
    '''
    Connect to the specifed device using the RPcoX driver

    Note that the appropriate RPcoX.Connect method is called so you do not need
    to perform that step in your code.

    Parameters
    ----------
    name : {'RZ6', 'RZ5', 'RP2', ... (any valid device string) }
        Name of device (as defined by the corresponding RPcoX.Connect* method).
    interface : {'GB', 'USB'}
        Type of interface (depends on the card that you have from TDT). See the
        TDT ActiveX documentation for clarification on which interface you
        would be using if you are still unsure.
    device_id : int (default 1)
        Id of device in the rack.  Only applicable if you have more than one of
        the same device (e.g. two RX6 devices).
    address : {None, (hostname, port)}
        If None, loads the ActiveX drivers directly, otherwise connects to the
        remote server specified by the hostname, port tuple.
    '''
    import pythoncom
    from . import actxobjects
    pythoncom.CoInitialize()
    debug_string = '%s %d via %s interface' % (name, device_id, interface)
    log.debug(debug_string)
    if address is None:
        driver = actxobjects.RPcoX()
    else:
        driver = dsp_server.RPcoXNET(address)
    if not getattr(driver, 'Connect%s' % name)(interface, device_id):
        raise DSPError(name, "Connection failed")
    log.debug("Connected to %s", name)
    return driver


def get_cof_path(circuit_name):
    '''
    Given relative path, returns absolute path to circuit file.  The *.rcx
    extension may be omitted.
    '''
    search_dirs = [os.path.join(os.path.dirname(__file__), 'components'),
                   os.getcwd(), ]
    log.debug("Searching %r", search_dirs)

    success = False
    if not circuit_name.endswith('.rcx'):
        circuit_name += '.rcx'

    log.debug("Attempting to locate circuit %s", circuit_name)
    for dir in search_dirs:
        circuit_path = os.path.join(dir, circuit_name)
        log.debug('Checking %s', circuit_path)
        if os.path.exists(circuit_path):
            success = True
            break

    if not success:
        raise IOError("Could not find circuit %s" % circuit_name)
    return circuit_path


def dtype_to_type_str(data_type):
    '''
    Convert Numpy dtype to the type string required by TDT's libraries

    TDT's ActiveX ReadTagVEX and WriteTagVEX functions require the type string
    to be one of I8, I16, I32 or F32.  Any valid format for specify Numpy dtype
    is supported.

    >>> dtype_to_type_str(np.int32)
    'I32'
    >>> dtype_to_type_str(np.float32)
    'F32'
    >>> dtype_to_type_str('float32')
    'F32'
    >>> dtype_to_type_str('int8')
    'I8'

    If a certain type is not supported by TDT, a Value error is raised:

    >>> dtype_to_type_str(np.float16)
    Traceback (most recent call last):
        ...
    ValueError: Unsupported Numpy dtype
    '''
    if np.issubdtype(data_type, np.integer):
        type_code = 'I'
    elif np.issubdtype(data_type, np.floating):
        type_code = 'F'
    else:
        raise ValueError("Unsupported Numpy dtype")

    # Since dtype.itemsize is the number of bytes, and the number in the TDT
    # type string reflects bit number, we can translate it by multiplying by 8.
    # Likewise, dtype.char is 'i' for integer and 'f' for floating point
    # datatypes.
    type_str = "{0}{1}".format(type_code, data_type.itemsize*8)
    log.debug("%r TDT type string is %s", data_type, type_str)
    if type_str not in ['F32', 'I32', 'I16', 'I8']:
        raise ValueError("Unsupported dtype")
    return type_str


def best_sf(data_type, range):
    '''
    Computes the optimal scaling factor for data compression

    Parameters
    ----------
    data_type
        Data type that values are being compressed to
    range : scalar or tuple
        Expected data range.  If scalar, assumes the value falls in the range
        (-range, range)
    '''
    data_type = np.dtype(data_type)
    try:
        info = np.iinfo(data_type)
    except:
        info = np.finfo(data_type)
    return info.max/np.abs(range).max()


def resolution(data_type, scaling_factor):
    '''
    Computes resolution for data type given scaling factor

    Parameters
    ----------
    data_type : dtype
        Numpy data type (or string)
    scaling_factor : float
        Scaling factor applied to data
    '''
    data_type = np.dtype(data_type)
    if np.issubdtype(data_type, np.integer):
        return 1/float(scaling_factor)
    else:
        raise ValueError("Float data types not supported")

CTYPES_TO_NP = {
    ctypes.c_char: np.int8,
    ctypes.c_wchar: np.int16,
    ctypes.c_byte: np.int8,
    ctypes.c_ubyte: np.uint8,
    ctypes.c_short: np.int16,
    ctypes.c_ushort: np.uint16,
    ctypes.c_int: np.int32,
    ctypes.c_uint: np.int32,
    ctypes.c_long: np.int32,
    ctypes.c_ulong: np.int32,
    ctypes.c_float: np.float32,
    ctypes.c_double: np.float64,
}
# Reverse lookup
NP_TO_CTYPES = dict((np.dtype(v), k) for k, v in CTYPES_TO_NP.items())


def shmem_as_ndarray(raw_array):
    '''
    Create a ndarray wrapper around shared memory space
    '''
    address = raw_array._wrapper.get_address()
    size = raw_array._wrapper.get_size()
    dtype = CTYPES_TO_NP[raw_array._type_]

    class NDArrayView(object):
        pass

    d = NDArrayView()
    d.__array_interface__ = {
        'data': (address, False),
        'typestr': np.dtype('uint8').str,
        'descr': np.dtype('uint8').descr,
        'shape': (size,),
        'strides': None,
        'version': 3,
    }
    return np.asarray(d).view(dtype=dtype)

if __name__ == '__main__':
    import doctest
    doctest.testmod()



from .dsp_process import DSPProcess
import time


def main():
    process = DSPProcess()
    circuit = process.load_circuit('components/test_physiology_RZ5', 'RZ5')
    # processed_buffer = circuit.get_buffer('processed', 'r', channels=16)
    raw_buffer = circuit.get_buffer('raw', 'r', channels=16, block_size=4098)
    print("BUFFER SIZE", raw_buffer.size)

    process.start()

    last_ctime = 0

    read = 0
    time.sleep(2)
    shape = raw_buffer.read().shape
    read += shape[-1]
    print(shape)
    ctime = circuit.get_tag('zTime')
    print(ctime, ctime-last_ctime)
    last_ctime = ctime
    time.sleep(2)
    shape = raw_buffer.read().shape
    read += shape[-1]
    print(shape)
    ctime = circuit.get_tag('zTime')
    print(ctime, ctime-last_ctime)
    last_ctime = ctime
    time.sleep(2)
    shape = raw_buffer.read().shape
    read += shape[-1]
    print(shape)
    ctime = circuit.get_tag('zTime')
    print(ctime, ctime-last_ctime)
    last_ctime = ctime
    time.sleep(2)
    shape = raw_buffer.read().shape
    read += shape[-1]
    print(shape)
    ctime = circuit.get_tag('zTime')
    print(ctime, ctime-last_ctime)
    last_ctime = ctime
    circuit.stop()
    time.sleep(2)
    shape = raw_buffer.read().shape
    read += shape[-1]
    print(shape)
    print("Samples read", read)
    ctime = circuit.get_tag('zTime')
    print(ctime, ctime-last_ctime)
    last_ctime = ctime
    process.terminate()

    # from pylab import *
    # print 'import'
    # comtypesplot(data[0])
    # print 'plotted'
    # show()
    # print 'show'
    # for i in range(10):
    #     time.sleep(0.1)
    #     print "PROC_READ", raw_buffer.read().shape
    # process.get_tag('ch1_out_sf')
    # process.set_tag('ch1_out_sf', 50e3)
    # process.get_tag('ch1_out_sf')
    # time.sleep(1)
    # process.terminate()

if __name__ == "__main__":
    main()
    # import cProfile
    # cProfile.run('main()')
    # import pstats
    # p = pstats.Stats('profile.dmp')
    # p.strip_dirs().sort_stats('cumulative').print_stats(50)



# There are apparently two more possible RCX types: 74 and 78.  It's not clear
# what these represent.

RCX_UNDEFINED = 65
RCX_BUFFER = 68
RCX_INTEGER = 73
RCX_FLOAT = 83
RCX_COEFFICIENT = 80
RCX_BOOL = 76

RCX_CAST = {
    RCX_INTEGER: int,
    RCX_FLOAT: float,
    RCX_BOOL: bool,
}

# Eventually I'd like to bind the device-specific variables as "constants" on
# the correct DSPCircuit so people can leverage these in their programs rather
# than having to look up the correct channel.  However, we need to compile a
# list of all the necessary module-level variables first.
RX6_DEFAULTS = {
    'DAC_CHANNEL_1':    1,
    'DAC_CHANNEL_2':    2,
    'ADC_CHANNEL_1':    128,
    'ADC_CHANNEL_2':    129,
}

# Value is the bit number containing the status for the corresponding value as
# returned by RP.GetStatus()
RCX_STATUS_BITMASK = {
    'connected':   0,
    'loaded':   1,
    'running':   2,
}



import numpy as np
import unittest

RZ6_ATTEN_BASE = 255*256*256


def split_attenuation(att):
    '''
    Split the attenuation value into two parts, one that can be realized by
    setting the hardware attenuator (the RZ6 attenuator can be set to 0, 20, 40
    or 60 dB) and the second that must be realized by scaling the waveform.

    Returns (hw_attenuation, sw_attenuation)
    ----------------------------------------
    hw_attenuation : float
        hardware attenuation setting
    sw_attenuation : float
        remaining attenuation that must be realized by scaling the waveform
    '''
    hw_step = min(int(att/20.0), 3)
    hw_att = hw_step*20.0
    sw_att = att-hw_att
    return hw_att, sw_att


def db_to_sf(db):
    sf = 10**(-db/20.0)
    return sf


def atten_to_bits(attA, attB, mute=False):
    '''
    Given the desired attenuation, determine the bit value that needs to be
    inserted into memory address 2 on the RZ6 to set the hardware attenuators
    accordingly.  This can replace the computations perfomed in the RZ6
    AudioOut macro which takes up to 12% of CPU time at 100 kHz.

    Raises ValueError if requested attenuation falls outside range of valid
    values.
    '''
    if not (0 <= attA <= 120):
        raise ValueError("Attenuation A must be between 0 and 120 dB")
    if not (0 <= attB <= 120):
        raise ValueError("Attenuation B must be between 0 and 120 dB")
    attA = min(int(attA/20), 3)
    attB = min(int(attB/20), 3) << 4
    mute = int(mute) * 256
    return RZ6_ATTEN_BASE | attA | attB | mute


def waveform_to_bits(waveform, sf):
    '''
    Scale and convert waveform to the binary data that needs to be inserted
    (via the RPvds poke component) into memory address 15 (for Out-A) or 16
    (for Out-B).  This replaces the computations performed in the RZ6 AudioOut
    macro which take up to 4% of CPU time at 100 kHz.  Used in combination with
    atten_to_bits, you can free up 16% of CPU time.
    '''
    waveform = np.asanyarray(waveform)
    waveform = waveform*sf*2e8
    waveform = waveform.astype('i')
    return (waveform | 255) ^ ((waveform >> 24) & 255)


class TestWithHardware(unittest.TestCase):

    CIRCUIT = '../components/test_RZ6_audio_out'

    def setUp(self):
        from tdt import DSPCircuit
        from os.path import dirname, join
        circuit = join(dirname(__file__), self.CIRCUIT)
        self.circuit = DSPCircuit(circuit, 'RZ6')
        self.buffer_in = self.circuit.get_buffer('in', 'w')
        self.buffer_out = self.circuit.get_buffer('out', 'r', src_type='int32',
                                                  dest_type='int32',
                                                  block_size=1)
        self.circuit.start()

    def test_waveform_to_bits(self):
        from numpy.random import random
        from numpy.testing import assert_array_equal  # noqa
        waveform = random(100e3).astype('float32')
        self.circuit.set_tag('nHi', 100e3)
        self.buffer_in.set(waveform)
        actual = self.buffer_out.acquire(1, 'running', False).ravel()
        expected = waveform_to_bits(waveform, 1)

        # Due to various vagaries of shuffling bits around, I believe the least
        # significant bit is sometimes lost when uploading to the RZ6.  This
        # results in a maximum difference between the actual and expected
        # values of 256.  It's not clear exactly what the significance of this
        # number actually is to me.
        difference = actual-expected
        self.assertTrue(difference.max() <= 256)


class TestRZ6Functions(unittest.TestCase):

    # attA, attB, expected bit value
    ATTEN_TEST_VALUES = [
        (0, 20, 16711696),
        (0, 40, 16711712),
        (0, 60, 16711728),
        (20, 0, 16711681),
        (20, 40, 16711713),
        (20, 60, 16711729),
        (40, 0, 16711682),
        (40, 20, 16711698),
        (40, 60, 16711730),
        (60, 0, 16711683),
        (60, 20, 16711699),
        (60, 40, 16711715),
    ]

    def test_attenuation_to_bits(self):
        for a, b, expected in self.ATTEN_TEST_VALUES:
            actual = atten_to_bits(a, b)
            self.assertEqual(actual, expected)

        # Ensure out of range attenuations raise an error
        self.assertRaises(ValueError, atten_to_bits, 0, 121)
        self.assertRaises(ValueError, atten_to_bits, 0, -1)
        self.assertRaises(ValueError, atten_to_bits, 121, 0)
        self.assertRaises(ValueError, atten_to_bits, -1, 0)

if __name__ == '__main__':
    unittest.main()



from os.path import abspath, split
import atexit
import numpy as np
import time

from .util import get_cof_path, connect_zbus, connect_rpcox
from .dsp_buffer import ReadableDSPBuffer, WriteableDSPBuffer
from .convert import convert
from .dsp_error import DSPError
from .constants import RCX_COEFFICIENT, RCX_CAST, RCX_STATUS_BITMASK

import logging
log = logging.getLogger(__name__)


class DSPCircuit:
    '''
    Wrapper around the TDT ActiveX object, RPCo.X.

    Provides several stringent checks and convenience methods to minimize
    programming errors and typos.

    circuit_name : string
        Path to circuit file.
    device_name : string
        Device to load circuit to.
    interface : {'GB', 'USB'}
        Interface to use (see TDT's ActiveX documentation on the Connect*
        methods for more information).  You almost certainly want 'GB' (which
        is the default value).
    device_id : number
        ID of device
    load : boolean (optional)
        Load circuit to specified device.  Default is True. Set to False if you
        just want to get a list of the tags available in the circuit.
    start : boolean (optional)
        Start (i.e. run) the circuit after loading it.  Default is False.
    address : two-tuple (str, int)
        Connect to the address specified as a two-tuple in (host, port) format
        using the network-aware proxy of TDT's driver.  If None, defaults to
        the TDT implementation of the RPcoX and zBUSx drivers.
    latch_trigger : {None, 1, 2, 3, 4}
        Trigger used for latching values when we need to capture a snapshot of
        some tags at a given point in time. This is used by DSPBuffer to
        eliminate race conditions when reading the value of the cycle and index
        tags. If the tags are not latched, then it's possible to read the index
        tag, then by the time the cycle tag is read the index has wrapped
        around to the beginning of the buffer.
    '''
    def __init__(self, circuit_name, device_name, interface='GB', device_id=1,
                 load=True, start=False, fs=None, address=None,
                 latch_trigger=None):
        self.device_name = device_name
        self.name = split(circuit_name)[1]
        self.device_id = device_id
        self.server_address = address
        self.interface = interface
        self.latch_trigger = latch_trigger

        # Hint for Matlab users: _iface is the same COM object a Matlab user
        # typically works with when they call actxserver('RPco.X').  It
        # supports the exact same methods as the Matlab version.
        self._iface = connect_rpcox(device_name, interface=interface,
                                    device_id=device_id, address=address)
        self._zbus = connect_zbus(interface=interface, address=address)
        self.path = get_cof_path(abspath(circuit_name))
        if load:
            self.load()
            atexit.register(self.stop)
        else:
            self.read()
        self.inspect()

        if start:
            self.start()

    def __getstate__(self):
        '''
        Provides support for pickling, which is required by the multiprocessing
        module for launching a new process.  _iface and _zbus are PyIDispatch
        objects which do not support pickling, so we just delete them and
        pickle the rest.
        '''
        state = self.__dict__.copy()
        del state['_iface']
        del state['_zbus']
        return state

    def __setstate__(self, state):
        '''
        Loads the state and reconnects the COM objects
        '''
        self.__dict__.update(state)
        self._iface = connect_rpcox(self.device_name)
        self._zbus = connect_zbus(self.server_address)
        self.read()

    def read(self):
        if not self._iface.ReadCOF(self.path):
            mesg = "Unable to read %s" % self.path
            raise DSPError(self, mesg)

    def load(self):
        '''
        Clear DSP RAM set all variables to default value

        The circuit is reloaded from disk, so any recent edits to the circuit
        will be reflected in the running program.
        '''
        # Note that ClearCOF alone does not apear to reset all variables to
        # default value, so we reload the circuit as well.
        self._iface.Halt()
        if not self._iface.ClearCOF():
            mesg = "Unable to clear %s buffers" % self
            raise DSPError(self, mesg)
        if not self._iface.LoadCOF(self.path):
            mesg = 'Unable to load %s' % self.path
            raise DSPError(self, mesg)
        log.debug("Reloaded %s", self)
        self.trigger('A', 'low')
        self.trigger('B', 'low')

    def inspect(self):
        '''
        Determine what tags are available in the microcode
        '''
        # Inspect COF for available tags and their type/size
        num_tags = self._iface.GetNumOf('ParTag')
        self.tags = {}
        self.scalar_tags = []
        self.vector_tags = []
        for i in range(num_tags):
            name = self._iface.GetNameOf('ParTag', i+1)
            if not name.startswith('%'):
                tag_size = self._iface.GetTagSize(name)
                if tag_size == 0:
                    mesg = "Attempted to initialize invalid tag %s" % name
                    raise DSPError(self, mesg)
                tag_type = self._iface.GetTagType(name)
                self.tags[name] = (tag_size, tag_type)
                if tag_size == 1:
                    self.scalar_tags.append(name)
                else:
                    self.vector_tags.append(name)
                log.debug("%s: found %s (size %d, type %s)", self, name,
                          tag_size, tag_type)
        self.fs = self._iface.GetSFreq()

    def cget_tag(self, name, tag_unit, val_unit):
        '''
        Enhanced version of `get_tag` that returns value in requested unit

        Parameters
        ----------
        name : str
            Tag name
        tag_unit : str
            Unit of tag
        val_unit : str
            Requested unit
        '''
        value = self.get_tag(name)
        return convert(tag_unit, val_unit, value, self.fs)

    def cset_tag(self, name, value, val_unit, tag_unit):
        '''
        Enhanced version of `set_tag` that converts the value

        Parameters
        ==========
        name : str
            Name of the parameter tag to write the converted value to
        value : int or float
            Value to convert
        val_unit : str
            Unit of value provided
        tag_unit : str
            Unit parameter tag requires

        Returns
        =======
        Actual value of the tag (i.e. the converted value)

        Value will be converted from val_unit to tag_unit based on the sampling
        frequency of the device (if needed).  See :module:`convert` for more
        information.
        '''
        value = convert(val_unit, tag_unit, value, self.fs)
        self.set_tag(name, value)
        return value

    def get_tag(self, name):
        '''
        Analogue of RPco.X.GetTagVal

        Parameters
        ==========
        name : str
            Name of the parameter tag to read the value from

        Raises DSPError
            If the tag does not exist or is not a scalar value (e.g. you cannot
            use this method with parameter tags linked to a buffer)
        '''
        # Check to see if tag exists
        try:
            tag_size, tag_type = self.tags[name]
        except KeyError:
            raise DSPError(self, "Tag %s does not exist" % name)
        if tag_size != 1:
            raise DSPError(self, "Tag %s is not a scalar value" % name)
        value = self._iface.GetTagVal(name)

        # The ActiveX wrapper always returns a float, regardless of whether
        # it's a bool, int or float.  Let's be sure to cast the value to the
        # correct type.
        if tag_type in RCX_CAST:
            value = RCX_CAST[tag_type](value)
        log.debug("Get %s:%s is %r", self, name, value)
        return value

    def set_tag(self, name, value):
        '''
        Analogue of RPco.X.SetTagVal

        Parameters
        ==========
        name : str
            Name of the parameter tag to write the value to
        value : int or float
            Value to write

        Raises DSPError
            If the tag does not exist or is not a scalar value (e.g. you cannot
            use this method with parameter tags linked to a buffer)
        '''
        # Check to see if tag exists
        if name not in self.tags:
            raise DSPError(self, "Tag %s does not exist" % name)
        if self.tags[name][0] != 1:
            raise DSPError(self, "Tag %s is not a scalar value" % name)
        if not self._iface.SetTagVal(name, float(value)):
            mesg = "Unable to set tag %s to %r" % (name, value)
            raise DSPError(self, mesg)
        log.debug("Set %s:%s to %r", self, name, value)

    def set_tags(self, **tags):
        '''
        Convenience function for setting the value of multiple tags

        >>> circuit.set_tags(record_duration=5, play_duration=4)
        '''
        for tag, value in tags.items():
            self.set_tag(tag, value)

    def set_coefficients(self, name, data):
        '''
        Load data to a coefficient or matrix input

        Parameters
        ==========
        name : str
            Name of the parameter tag to write the data to
        data : array-like
            Data to write to tag.  Must be 1D format (even for matrices).
            See RPvds documentation for appropriate ordering of indices for
            the component.

        Raises DSPError
            If the specified parameter tag is not linked to a coefficient input
            or the length of the data is not equal to the size of the input on
            the component.

        Note that as of 3.10.2011, RPvds' CoefLoad component appears to be
        broken (per conversation with TDT's tech support -- Mark Hanus and
        Chris Walters).  As a workaround, connect a data tag directly to the
        >K or >Coef input of the component.
        '''
        tag_size, tag_type = self.tags[name]
        if tag_type != RCX_COEFFICIENT:
            raise DSPError(self, "Tag %s is not a coefficient buffer" % name)
        if tag_size != len(data):
            mesg = "Exactly %d values must be written to tag %s"
            raise DSPError(self, mesg % (tag_size, name))
        if not self._iface.WriteTagV(name, 0, data):
            raise DSPError(self, "Unable to upload data to buffer %s" % name)

    def start(self, pause=0.25):
        '''
        Analogue of RPco.X.Run

        The circuit sometimes requires a couple hundred msec "settle" before we
        can commence data acquisition
        '''
        log.debug("starting %s", self)
        if not self._iface.Run():
            raise DSPError(self, "Unable to start circuit")
        time.sleep(pause)

    def stop(self):
        '''
        Analogue of RPco.X.Halt
        '''
        log.debug("stopping %s", self)
        if not self._iface.Halt():
            raise DSPError(self, "Unable to stop circuit")

    def trigger(self, trigger, mode='pulse'):
        '''
        Fire a zBUS or software trigger

        Parameters
        ----------
        trigger : {1-9, 'A', 'B'}
            Fire the specified trigger.  If integer, this corresponds to
            RPco.X.SoftTrg.  If 'A' or 'B', this fires the corresponding zBUS
            trigger.
        mode : {'pulse', 'high', 'low'}
            Relevant only when trigger is 'A' or 'B'.  Indicates the
            corresponding mode to set the zBUS trigger to

        Note that due to a bug in the TDT ActiveX library for versions greater
        than 56, we have no way of ensuring that zBUS trigger A or B were
        fired.
        '''
        # Convert mode string to the corresponding integer
        mode_enum = dict(pulse=0, high=1, low=2)
        # We have no way of ensuring that zBUS trigger A or B were fired
        # properly due to a bug in versions of the ActiveX library >= 57.
        if trigger == 'A':
            self._zbus.zBusTrigA(0, mode_enum[mode], 10)
        elif trigger == 'B':
            self._zbus.zBusTrigB(0, mode_enum[mode], 10)
        elif (1 <= trigger < 10) and mode == 'pulse':
            if not self._iface.SoftTrg(trigger):
                raise DSPError(self, "Could not fire soft trigger %d"
                               % trigger)
        else:
            mesg = "Unsupported trigger mode %s %s" % (trigger, mode)
            raise DSPError(self, mesg)
        log.debug('Trigger %r %s', trigger, mode)

    def get_buffer(self, data_tag, mode, *args, **kw):
        kw.setdefault('latch_trigger', self.latch_trigger)
        if mode == 'w':
            return WriteableDSPBuffer(self, data_tag, *args, **kw)
        elif mode == 'r':
            return ReadableDSPBuffer(self, data_tag, *args, **kw)

    def clear_buffer(self, name):
        if not self._iface.ZeroTag(name):
            raise DSPError(self, "Unable to zero tag %s" % name)

    def get_status(self, status):
        return (self._iface.GetStatus() >> RCX_STATUS_BITMASK[status]) & 1

    def is_connected(self):
        '''
        True if connection with hardware is active, False otherwise
        '''
        return self.get_status('connected')

    def is_loaded(self):
        '''
        True if microcode is loaded, False otherwise
        '''
        return self.get_status('loaded')

    def is_running(self):
        return self.get_status('running')

    def __str__(self):
        state = ''
        if self.is_connected():
            state += 'Cx'
        if self.is_loaded():
            state += 'Ld'
        if self.is_running():
            state += 'Rn'
        return "{}:{}:{}".format(self.device_name, self.name, state)

    def set_sort_windows(self, name, windows):
        '''
        Utility function for configuring TDT's SpikeSort component coefficients

        Windows should be a list of 3-tuples in the format (time, center volt,
        half-height)

        If the windows overlap in time such that they cannot be converted into
        a coefficient buffer, and error will be raised.
        '''
        tag_size, tag_type = self.tags[name]
        if tag_type != RCX_COEFFICIENT:
            raise DSPError(self, "Tag %s is not a coefficient buffer" % name)
        coefficients = np.zeros(tag_size).reshape((-1, 3))
        processed = []
        for i, window in enumerate(windows):
            t, center, height = window
            x = round(t*self.fs)
            if x in processed:
                raise DSPError(self, "Windows for %s overlap in time" % name)
            processed.append(x)
            coefficients[x] = center, height, i+1
        self.set_coefficients(name, coefficients.ravel())

    def convert(self, value, src_unit, dest_unit):
        '''
        Converts value to desired unit give the sampling frequency of the DSP.

        Parameters specified in paradigms are typically expressed as frequency
        and time while many DSP parameters are expressed in number of samples
        (referenced to the DSP sampling frequency).  This function provides a
        convenience method for converting between conventional values and the
        'digital' values used by the DSP.

        Note that for converting units of time/frequency to n/nPer, we have to
        coerce the value to a multiple of the DSP period (e.g. the number of
        'ticks' of the DSP clock).

        Appropriate strings for the unit types:

            fs
                sampling frequency
            nPer
                number of samples per period
            n
                number of samples
            s
                seconds
            ms
                milliseconds
            nPow2
                number of samples, coerced to the next greater power of 2 (used
                for ensuring efficient FFT computation)

        Given a DSP clock frequency of 10 kHz::

            >>> circuit.convert(0.5, 's', 'n')
            5000
            >>> circuit.convert(500, 'fs', 'nPer')
            20

        Given a DSP clock frequency of 97.5 kHz::

            >>> circuit.convert(5, 's', 'nPow2')
            524288

        Parameters
        ----------
        value: numerical (e.g. integer or float)
            Value to be converted
        src_unit: string
            Unit of the value
        dest_unit: string
            Destination unit

        Returns
        -------
        converted unit : numerical value
        '''
        return convert(src_unit, dest_unit, value, self.fs)

    def print_tag_info(self):
        '''
        Prints a list of tags and their current value if they are a scalar
        (buffer tags are not printed yet)

        Used as a convenience method for debugging
        '''
        for tag, (tag_size, tag_type) in self.tags.items():
            if tag_size == 1:
                print(tag, self.get_tag(tag))



def ispow2(n):
    '''
    True if n is a power of 2, False otherwise

    >>> ispow2(5)
    False
    >>> ispow2(4)
    True
    '''
    return (n & (n-1)) == 0


def nextpow2(n):
    '''
    Given n, return the nearest power of two that is >= n

    >>> nextpow2(1)
    1
    >>> nextpow2(2)
    2
    >>> nextpow2(5)
    8
    >>> nextpow2(17)
    32
    '''
    if ispow2(n):
        return n

    count = 0
    while n != 0:
        n = n >> 1
        count += 1
    return 1 << count


class SamplingRateError(ValueError):
    '''
    Indicates that the conversion of frequency to sampling rate could not be
    performed.
    '''

    def __init__(self, fs, requested_fs):
        self.fs = fs
        self.requested_fs = requested_fs

    def __str__(self):
        mesg = 'The requested sampling rate, %f Hz, is greater than ' + \
               'the DSP clock frequency of %f Hz.'
        return mesg % (self.requested_fs, self.fs)


def convert(src_unit, dest_unit, value, dsp_fs):
    '''
    Converts value to desired unit give the sampling frequency of the DSP.

    Parameters specified in paradigms are typically expressed as
    frequency and time while many DSP parameters are expressed in number of
    samples (referenced to the DSP sampling frequency).  This function provides
    a convenience method for converting between conventional values and the
    'digital' values used by the DSP.

    Note that for converting units of time/frequency to n/nPer, we have to
    coerce the value to a multiple of the DSP period (e.g. the number of
    'ticks' of the DSP clock).

    Appropriate strings for the unit types:

        fs
            sampling frequency
        nPer
            number of samples per period
        n
            number of samples
        s
            seconds
        ms
            milliseconds
        nPow2
            number of samples, coerced to the next greater power of 2 (used for
            ensuring efficient FFT computation)

    >>> convert('s', 'n', 0.5, 10000)
    5000
    >>> convert('fs', 'nPer', 500, 10000)
    20
    >>> convert('s', 'nPow2', 5, 97.5e3)
    524288

    Parameters
    ----------
    src_unit: string
    dest_unit: string
        Destination unit
    value: numerical (e.g. integer or float)
        Value to be converted

    Returns
    -------
    converted unit : numerical value
    '''
    def fs_to_nPer(req_fs, dsp_fs):
        if dsp_fs < req_fs:
            raise SamplingRateError(dsp_fs, req_fs)
        return int(dsp_fs/req_fs)

    def nPer_to_fs(nPer, dsp_fs):
        return dsp_fs/nPer

    def n_to_s(n, dsp_fs):
        return n/dsp_fs

    def s_to_n(s, dsp_fs):
        return int(s*dsp_fs)

    def ms_to_n(ms, dsp_fs):
        return int(ms*1e-3*dsp_fs)

    def n_to_ms(n, dsp_fs):
        return n/dsp_fs*1e3

    def s_to_nPow2(s, dsp_fs):
        return nextpow2(s_to_n(s, dsp_fs))

    fun = '%s_to_%s' % (src_unit, dest_unit)
    return locals()[fun](value, dsp_fs)



import numpy as np
import time
import logging
import unittest

from .dsp_circuit import DSPCircuit

logging.basicConfig(level=logging.DEBUG)


# class TestPickling(unittest.TestCase):
#
#    def setUp(self):
#        self.circuit_file = 'test_circuit.tmp'
#        self.buffer_file = 'test_buffer.tmp'
#        self.circuit = DSPCircuit('components/test_physiology_RZ6', 'RZ6')
#        self.buffer = self.circuit.get_buffer('raw', 'r')
#        with open(self.circuit_file, 'w') as fh:
#            pickle.dump(self.circuit, fh)
#        with open(self.buffer_file, 'w') as fh:
#            pickle.dump(self.buffer, fh)
#
#    def test_persist(self):
#        with open(self.circuit_file, 'r') as fh:
#            circuit = pickle.load(fh)
#        with open(self.buffer_file, 'r') as fh:
#            buffer = pickle.load(fh)
#
#    def tearDown(self):
#        remove(self.circuit_file)
#        remove(self.buffer_file)
#


class TestBufferRead(unittest.TestCase):

    def setUp(self):
        self.circuit = DSPCircuit('components/data_reduction', 'RZ6')
        self.circuit.set_tag('nHi', 4)
        self.circuit.start()
        self.shape = (4, 4)

    def tearDown(self):
        self.circuit.stop()

    def assertShape(self, data):
        self.assertEquals(data.shape, self.shape)

    def assertSeqSamples(self, data, res):
        seq_samples = np.abs(np.diff(data, axis=1)-0.1) < res
        self.assertTrue(seq_samples.all())

    def assertSeqChannels(self, data, res):
        ch_samples = np.abs(np.diff(data, axis=0)-1) < res
        self.assertTrue(ch_samples.all())

    def assertValid(self, buffer):
        self.circuit.trigger(1)
        time.sleep(0.1)
        data = buffer.read()
        self.assertShape(data)
        self.assertSeqSamples(data, buffer.resolution)
        if data.ndim > 1:
            self.assertSeqChannels(data, buffer.resolution)

    # def test_overflow_warning(self):
    #    buffer = self.circuit.get_buffer('mc', src_type='float32', channels=4,
    #            block_size=1)
    #    self.circuit.set_tag('nHi_nodata', buffer.size+100)
    #    self.circuit.trigger(2)
    #    while self.circuit.get_tag('t2'):
    #        import time
    #        time.sleep(1)
    #    self.assertRaises(DSPError, buffer.read)

    def test_mc2(self):
        buffer = self.circuit.get_buffer('mc', 'r', src_type='float32',
                                         channels=4, block_size=4)
        self.circuit.trigger(2)
        self.circuit.trigger(2)
        data = buffer.read()
        self.assertSeqChannels(data, buffer.resolution)

        self.circuit.trigger(1)
        data = buffer.read()
        self.assertSeqChannels(data, buffer.resolution)
        self.assertSeqSamples(data, buffer.resolution)

    def test_mc(self):
        buffer = self.circuit.get_buffer('mc', 'r', src_type='float32',
                                         channels=4, block_size=1)
        self.assertValid(buffer)

    def test_sh8(self):
        buffer = self.circuit.get_buffer('sh8', 'r', src_type='int8',
                                         channels=4, block_size=4)
        self.assertEqual(self.circuit.get_tag('sh8_sf'), 31)
        self.assertValid(buffer)

    def test_sh16(self):
        buffer = self.circuit.get_buffer('sh16', 'r', src_type='int16',
                                         channels=2, block_size=4)
        self.assertEqual(self.circuit.get_tag('sh16_sf'), 6553)
        self.shape = (2, 4)
        self.assertValid(buffer)

    def test_c16(self):
        # If this fails, double check that scaling factor is set properly in
        # the circuit since I am testing the ability to use a different tag for
        # computing the scaling value
        buffer = self.circuit.get_buffer('c16', 'r', src_type='int16',
                                         sf_tag='sh16_sf', block_size=4)
        self.shape = (1, 4)
        self.assertValid(buffer)

    def test_mcFI16(self):
        buffer = self.circuit.get_buffer('mcFI16', 'r', src_type='int16',
                                         channels=8, block_size=4)
        self.shape = (8, 4)
        self.assertValid(buffer)

    def test_c8D(self):
        buffer = self.circuit.get_buffer('c8D', 'r', src_type='int8',
                                         block_size=4)
        self.shape = (1, 4)
        self.assertValid(buffer)

    # THIS KEEPS FAILING!
    # def test_mcFI8(self):
    #    buffer = self.circuit.get_buffer('mcFI8', src_type='int8',
    #                                     block_size=4, channels=16)
    #    self.shape = (16, 4)
    #    self.assertValid(buffer)


class TestBufferWrite(unittest.TestCase):

    def setUp(self):
        self.circuit = DSPCircuit('components/test_physiology_RZ6', 'RZ6')
        self.buffer = self.circuit.get_buffer('speaker', 'w')
        self.circuit.start()

    def test_zero(self):
        self.buffer.clear()

    def test_write(self):
        from numpy import sin, arange, random
        from numpy.testing import assert_array_almost_equal

        t = arange(10e3)/1e3
        data_a = sin(2*np.pi*10*t)
        self.buffer.write(data_a, force=True)
        data_b = random.normal(size=10e3)
        self.buffer.write(data_b)
        # n = len(data_a)
        written_a = self.buffer._iface.ReadTagV('speaker', 0, len(data_a))
        written_b = self.buffer._iface.ReadTagV('speaker', len(data_a),
                                                len(data_b))
        assert_array_almost_equal(written_a, data_a)
        assert_array_almost_equal(written_b, data_b)

    def tearDown(self):
        self.circuit.stop()


# class TestCOMWrapper(unittest.TestCase):
#
#    #def test_comtypes(self):
#    #    from comtypes import client
#    #    iface = client.CreateObject('RPco.X')
#    #    self.assertWrite(iface)
#    #    self.assertWrite(iface)
#
#    def test_win32com(self):
#        from win32com import client
#        iface = client.Dispatch('RPco.X')
#        self.assertWrite(iface)
#        self.assertWrite(iface)
#
#    def test_custom_actxobject(self):
#        from actxobjects import RPcoX
#        iface = RPcoX()
#        self.assertWrite(iface)
#        self.assertWrite(iface)
#
#    def assertWrite(self, iface):
#        from numpy import random
#        from numpy.testing import assert_array_almost_equal
#        from array import array
#        data = random.normal(size=100e3)
#        data = array('d', data)
#        iface.ConnectRZ6('GB', 1)
#        iface.LoadCOF('components/test_physiology_RZ6.rcx')
#        iface.Run()
#        result = iface.WriteTagV('speaker', 0, data)
#        self.assertEqual(result, 1)
#
#        written = iface.ReadTagV('speaker', 0, len(data))
#        assert_array_almost_equal(written, data)


# class TestProcess(unittest.TestCase):
#
#    def test_read_write(self):
#        write_size = 50000
#        total_write = 500000
#        process = DSPProcess()
#        circuit = process.load_circuit('components/test_read', 'RZ6')
#        write_buffer = circuit.get_buffer('write', 'w')
#        read_buffer = circuit.get_buffer('read', 'r')
#
#        process.start()
#        samples_written = []
#        samples_read = []
#        #available = write_buffer.available()
#
#        samples = np.random.uniform(size=write_size)
#        write_buffer.write(samples, timeout=1)
#        samples_written.extend(samples)
#        circuit.trigger('A', 'high')
#
#        self.assertEqual(write_buffer.total_samples_written, write_size)
#
#        while True:
#            if write_buffer.total_samples_written >= total_write:
#                break
#            samples = read_buffer.read()
#            samples_read.extend(samples[0])
#            samples = np.random.uniform(size=write_size)
#            write_buffer.write(samples, timeout=None)
#            samples_written.extend(samples)
#
#        total_samples = write_buffer.total_samples_written
#        while read_buffer.total_samples_read <= total_samples:
#            if read_buffer.pending():
#                samples = read_buffer.read()
#                samples_read.extend(samples[0])
#        process.stop()
#
#        written = np.array(samples_written)
#        read = np.array(samples_read)
#        #from pylab import *
#        #plot(np.arange(len(written)), written, 'k')
#        #plot(np.arange(len(read)), read+1, 'r')
#        #plot(np.arange(len(written)), written-read[:total_samples]+2, 'g')
#        #show()
#        np.testing.assert_almost_equal(written, read[:total_samples])
#
if __name__ == '__main__':
    unittest.main()



import logging
log = logging.getLogger(__name__)

import numpy as np


def wrap(length, offset, buffer_size):
    if (offset+length) > buffer_size:
        a_length = buffer_size-offset
        b_length = length-a_length
        return ((offset, a_length), (0, b_length))
    else:
        return ((offset, length), )


def span(old_cycle, old_idx, new_cycle, new_idx, buffer_size):
    '''
    Returns the number of slots between new_idx and old_idx given buffer size.
    '''
    start = int((old_cycle * buffer_size) + old_idx)
    end = int((new_cycle * buffer_size) + new_idx)
    if (start > end):
        log.info('Start sample %d (C %d, I %d), end sample %d (C %d, I %d), buffer %d',
                 start, old_cycle, old_idx, end, new_cycle, new_idx, buffer_size)
        raise ValueError('Start sample higher than end sample')
    if (end - start) > buffer_size:
        log.info('Start sample %d (C %d, I %d), end sample %d (C %d, I %d), buffer %d',
                 start, old_cycle, old_idx, end, new_cycle, new_idx, buffer_size)
        raise ValueError('Number of slots exceeds buffer size')
    return end-start


class AbstractRingBuffer:
    '''
    Subclasses must provide the following attributes (either by setting them in
    the __init__ method or providing property getter/setters):

        * read_index
        * write_index
        * read_cycle
        * write_cycle
        * size
        * channels
        * block_size
        * total_samples_written
        * total_samples_read

    Also, provide implementation of the following methods:
        * _read(self, offset, length)
        * _write(self, offset, data)
    '''

    def _offset_to_index(self, offset):
        if offset is None:
            offset = self.total_samples_written
        write_cycle, write_index = divmod(offset, self.size)
        log.debug('Offset %d is at cycle %d, index %d', offset, write_cycle,
                  write_index)
        log.debug('Current offset is cycle %d, index %d', self.write_cycle,
                  self.write_index)
        if self.write_cycle < write_cycle:
            raise ValueError('Offset too far back in time')
        return write_cycle, write_index

    def pending(self):
        '''
        Number of filled slots waiting to be read

        '''
        self.latch()
        return span(self.read_cycle, self.read_index, self.write_cycle,
                    self.write_index, self.size)

    def blocks_pending(self):
        '''
        Number of filled blocks waiting to be read
        '''
        return int(self.pending()/self.block_size)*self.block_size

    def available(self, offset=None):
        '''
        Number of empty slots available for writing

        Parameters
        ----------
        offset : {None, int}
            If specified, return number of samples relative to offset. Offset
            is relative to beginning of acquisition.
        '''
        self.latch()
        write_cycle, write_index = self._offset_to_index(offset)
        if (self.total_samples_written == 0) and (self.read_index == 0):
            return self.size
        log.debug('Available: write cycle %d index %d, '
                  'read cycle %d index %d, size %d',
                  write_cycle, write_index, self.read_cycle, self.read_index,
                  self.size)
        return self.size - \
            span(self.read_cycle, self.read_index,
                    write_cycle, write_index, self.size)

    def blocks_available(self):
        return int(self.available()/self.block_size)*self.block_size

    def read_all(self):
        return self.read(self.pending())

    def read(self, samples=None):
        '''
        Parameters
        ----------
        samples : int
            Number of samples to read. If None, read all samples acquired since
            last call to read.
        '''
        try:
            if samples is None:
                samples = self.blocks_pending()
            elif samples > self.pending():
                mesg = 'Attempt to read %r samples failed because only ' + \
                    '%r slots are available for read'
                raise ValueError(mesg % (samples, self.pending()))
        except ValueError:
            raise IOError('Read was too slow and unread samples were overwritten')

        data = [self._read(o, l) for o, l in wrap(samples, self.read_index, self.size)]
        data = np.concatenate(data, axis=-1)
        self.total_samples_read += samples
        self.read_cycle, self.read_index = divmod(self.total_samples_read, self.size)
        return data

    def write(self, data, offset=None):
        write_cycle, write_index = self._offset_to_index(offset)
        try:
            available = self.available(offset)
        except ValueError:
            raise IOError('Write was too slow and old samples were regenerated')
        samples = data.shape[-1]
        log.debug('Current write cycle %d and index %d with %d samples available to write',
                  self.write_cycle, self.write_index, available)

        if samples == 0:
            return
        elif samples > available:
            mesg = 'Attempt to write %d samples failed because only ' + \
                   '%d slots are available for write'
            raise ValueError(mesg % (samples, available))

        samples_written = 0
        for i, (o, l) in enumerate(wrap(samples, write_index, self.size)):
            lb = samples_written
            ub = samples_written + l
            if not self._write(o, data[..., lb:ub]):
                raise SystemError('Problem with writing data to buffer')
            samples_written += l

        if offset is not None:
            self.total_samples_written = offset + samples_written
        else:
            self.total_samples_written += samples_written
        self.write_cycle, self.write_index = \
            divmod(self.total_samples_written, self.size)
        log.debug('Write %s samples. Write pointer at %d cycles, %d index.',
                  samples_written, self.write_cycle, self.write_index)
        return samples_written

    def reset_read(self, index=None):
        '''
        Reset the read index
        '''
        if index is None:
            index = self._iface.GetTagVal(self.idx_tag)
        self.read_index = index



import itertools

from .. import DSPCircuit
from ..device.RZ6 import split_atten, atten_to_bits

if __name__ == '__main__':
    circuit = DSPCircuit('debug_RZ6_audio_out', 'RZ6')
    circuit.start()

    for a, b in itertools.permutations((0, 20, 40, 60), 2):
        circuit.set_tag('attA', a)
        circuit.set_tag('attB', b)
        padA = circuit.get_tag('Pad-A')
        padB = circuit.get_tag('Pad-B')
        poke = circuit.get_tag('Poke')
        guess = atten_to_bits(a, b)
        circuit.set_tag('test_val', guess)
        print('')
        print(a, b, padA, padB, poke, guess)
        # raw_input("Please check value on RZ6 LCD display")

    for a in (0, 12.5, 30, 40, 89.1):
        circuit.set_tag('attA', a)
        expected_sf = circuit.get_tag('SF-A')
        actual_att, actual_sf = split_atten(a)
        print(a, expected_sf, actual_sf)



# -*- coding: utf-8 -*-
# Created by makepy.py version 0.4.97
# By python version 2.5.2 (r252:60911, Feb 21 2008, 13:11:45) [MSC v.1310 32 bit (Intel)]
# From type library 'RPcoX.ocx'
# On Thu Aug 20 14:23:15 2009
"""RPcoX ActiveX Control module"""
makepy_version = '0.4.97'
python_version = 0x20502f0

import win32com.client.CLSIDToClass, pythoncom
import win32com.client.util
from pywintypes import IID
from win32com.client import Dispatch

# The following 3 lines may need tweaking for the particular server
# Candidates are pythoncom.Missing, .Empty and .ArgNotFound
defaultNamedOptArg=pythoncom.Empty
defaultNamedNotOptArg=pythoncom.Empty
defaultUnnamedArg=pythoncom.Empty

CLSID = IID('{D323A622-1D13-11D4-8858-444553540000}')
MajorVersion = 1
MinorVersion = 0
LibraryFlags = 10
LCID = 0x0

from win32com.client import DispatchBaseClass
class _DRPcoX(DispatchBaseClass):
    """Dispatch interface for RPcoX Control"""
    CLSID = IID('{D323A623-1D13-11D4-8858-444553540000}')
    coclass_clsid = IID('{D323A625-1D13-11D4-8858-444553540000}')

    def AboutBox(self):
        return self._oleobj_.InvokeTypes(-552, LCID, 1, (24, 0), (),)

    def ClearCOF(self):
        return self._oleobj_.InvokeTypes(28, LCID, 1, (3, 0), (),)

    def Connect(self, Interface=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(2, LCID, 1, (3, 0), ((3, 0), (3, 0)),Interface
            , DevNum)

    def ConnectRA16(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(24, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRL2(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(23, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRM1(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        """method ConnectRM1"""
        return self._oleobj_.InvokeTypes(38, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRM2(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        """method ConnectRM2"""
        return self._oleobj_.InvokeTypes(39, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRP2(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(22, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRV8(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(32, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRX5(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        """method ConnectRX5"""
        return self._oleobj_.InvokeTypes(40, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRX6(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        """method ConnectRX6"""
        return self._oleobj_.InvokeTypes(41, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRX7(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        """method ConnectRX7"""
        return self._oleobj_.InvokeTypes(42, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRX8(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        """method ConnectRX8"""
        return self._oleobj_.InvokeTypes(43, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRX9(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        """method ConnectRX9"""
        return self._oleobj_.InvokeTypes(44, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRZ2(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        """method ConnectRZ2"""
        return self._oleobj_.InvokeTypes(45, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRZ3(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        """method ConnectRZ3"""
        return self._oleobj_.InvokeTypes(46, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRZ4(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        """method ConnectRZ4"""
        return self._oleobj_.InvokeTypes(47, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRZ5(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        """method ConnectRZ5"""
        return self._oleobj_.InvokeTypes(48, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def ConnectRZ6(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
        """method ConnectRZ6"""
        return self._oleobj_.InvokeTypes(49, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
            , DevNum)

    def DefStatus(self, DefID=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(36, LCID, 1, (3, 0), ((3, 0),),DefID
            )

    def GetCycUse(self):
        return self._oleobj_.InvokeTypes(27, LCID, 1, (3, 0), (),)

    def GetDefData(self, DefID=defaultNamedNotOptArg):
        return self._ApplyTypes_(37, 1, (12, 0), ((3, 0),), u'GetDefData', None,DefID
            )

    def GetDevCfg(self, Addr=defaultNamedNotOptArg, Width32=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(33, LCID, 1, (3, 0), ((3, 0), (3, 0)),Addr
            , Width32)

    def GetError(self):
        # Result is a Unicode object - return as-is for this version of Python
        return self._oleobj_.InvokeTypes(1, LCID, 1, (8, 0), (),)

    def GetNameOf(self, ObjTypeName=defaultNamedNotOptArg, Index=defaultNamedNotOptArg):
        # Result is a Unicode object - return as-is for this version of Python
        return self._oleobj_.InvokeTypes(20, LCID, 1, (8, 0), ((8, 0), (3, 0)),ObjTypeName
            , Index)

    def GetNames(self, NameList=defaultNamedNotOptArg, MaxNames=defaultNamedNotOptArg, ObjType=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(13, LCID, 1, (3, 0), ((8, 0), (3, 0), (3, 0)),NameList
            , MaxNames, ObjType)

    def GetNumOf(self, ObjTypeName=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(19, LCID, 1, (3, 0), ((8, 0),),ObjTypeName
            )

    def GetSFreq(self):
        return self._oleobj_.InvokeTypes(31, LCID, 1, (4, 0), (),)

    def GetStatus(self):
        return self._oleobj_.InvokeTypes(26, LCID, 1, (3, 0), (),)

    def GetTagSize(self, Name=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(16, LCID, 1, (3, 0), ((8, 0),),Name
            )

    def GetTagType(self, Name=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(17, LCID, 1, (3, 0), ((8, 0),),Name
            )

    def GetTagVal(self, Name=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(8, LCID, 1, (4, 0), ((8, 0),),Name
            )

    def Halt(self):
        return self._oleobj_.InvokeTypes(6, LCID, 1, (3, 0), (),)

    def LoadCOF(self, FileName=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(4, LCID, 1, (3, 0), ((8, 0),),FileName
            )

    def LoadCOFsf(self, FileName=defaultNamedNotOptArg, SampFreq=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(35, LCID, 1, (3, 0), ((8, 0), (4, 0)),FileName
            , SampFreq)

    def ReadCOF(self, FileName=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(21, LCID, 1, (3, 0), ((8, 0),),FileName
            )

    def ReadTag(self, Name=defaultNamedNotOptArg, pBuf=defaultNamedNotOptArg, nOS=defaultNamedNotOptArg, nWords=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(9, LCID, 1, (3, 0), ((8, 0), (16388, 0), (3, 0), (3, 0)),Name
            , pBuf, nOS, nWords)
         # Hex tag is 0x44?  We want it to be of type VTS_PR4

    # For ReadTagV and ReadTagVEX, the default output of makepy is to use
    # _ApplyTypes_ to pass the data through _get_good_object_.  This function
    # slows down the I/O significantly, so we just call InvokeTypes directly.
    #
    # This is a hand-edited version of the makepy output.

    def ReadTagV(self, Name=defaultNamedNotOptArg, nOS=defaultNamedNotOptArg, nWords=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(14, 0, 1, (12, 0), ((8, 0), (3, 0), (3, 0)), Name , nOS, nWords)

    def ReadTagVEX(self, Name=defaultNamedNotOptArg, nOS=defaultNamedNotOptArg, nWords=defaultNamedNotOptArg, SrcType=defaultNamedNotOptArg
            , DstType=defaultNamedNotOptArg, nChans=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(25, 0, 1, (12, 0), ((8, 0), (3, 0), (3, 0), (8, 0), (8, 0), (3, 0)), Name
            , nOS, nWords, SrcType, DstType, nChans
            )

    def Run(self):
        return self._oleobj_.InvokeTypes(5, LCID, 1, (3, 0), (),)

    def SendParTable(self, Name=defaultNamedNotOptArg, IndexID=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(11, LCID, 1, (3, 0), ((8, 0), (4, 0)),Name
            , IndexID)

    def SendSrcFile(self, Name=defaultNamedNotOptArg, SeekOS=defaultNamedNotOptArg, nWords=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(12, LCID, 1, (3, 0), ((8, 0), (3, 0), (3, 0)),Name
            , SeekOS, nWords)

    def SetDevCfg(self, Addr=defaultNamedNotOptArg, Val=defaultNamedNotOptArg, Width32=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(34, LCID, 1, (3, 0), ((3, 0), (3, 0), (3, 0)),Addr
            , Val, Width32)

    def SetSrcFileName(self, Name=defaultNamedNotOptArg, FileName=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(18, LCID, 1, (3, 0), ((8, 0), (8, 0)),Name
            , FileName)

    def SetTagVal(self, Name=defaultNamedNotOptArg, Val=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(3, LCID, 1, (3, 0), ((8, 0), (4, 0)),Name
            , Val)

    def SoftTrg(self, Trg_Bitn=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(7, LCID, 1, (3, 0), ((3, 0),),Trg_Bitn
            )

    def WriteTag(self, Name=defaultNamedNotOptArg, pBuf=defaultNamedNotOptArg, nOS=defaultNamedNotOptArg, nWords=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(10, LCID, 1, (3, 0), ((8, 0), (16388, 0), (3, 0), (3, 0)),Name
            , pBuf, nOS, nWords)

    def WriteTagV(self, Name=defaultNamedNotOptArg, nOS=defaultNamedNotOptArg, Buf=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(15, LCID, 1, (3, 0), ((8, 0), (3, 0), (0x2005, 0)),Name
            , nOS, Buf)

    def WriteTagVEX(self, Name=defaultNamedNotOptArg, nOS=defaultNamedNotOptArg, DstType=defaultNamedNotOptArg, Buf=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(29, LCID, 1, (3, 0), ((8, 0), (3, 0), (8, 0), (0x2005, 0)),Name
            , nOS, DstType, Buf)

    def ZeroTag(self, Name=defaultNamedNotOptArg):
        return self._oleobj_.InvokeTypes(30, LCID, 1, (3, 0), ((8, 0),),Name
            )

    _prop_map_get_ = {
    }
    _prop_map_put_ = {
    }

class _DRPcoXEvents:
    """Event interface for RPcoX Control"""
    CLSID = CLSID_Sink = IID('{D323A624-1D13-11D4-8858-444553540000}')
    coclass_clsid = IID('{D323A625-1D13-11D4-8858-444553540000}')
    _public_methods_ = [] # For COM Server support
    _dispid_to_func_ = {
                1 : "OnDefComplete",
        }

    def __init__(self, oobj = None):
        if oobj is None:
            self._olecp = None
        else:
            import win32com.server.util
            from win32com.server.policy import EventHandlerPolicy
            cpc=oobj._oleobj_.QueryInterface(pythoncom.IID_IConnectionPointContainer)
            cp=cpc.FindConnectionPoint(self.CLSID_Sink)
            cookie=cp.Advise(win32com.server.util.wrap(self, usePolicy=EventHandlerPolicy))
            self._olecp,self._olecp_cookie = cp,cookie
    def __del__(self):
        try:
            self.close()
        except pythoncom.com_error:
            pass
    def close(self):
        if self._olecp is not None:
            cp,cookie,self._olecp,self._olecp_cookie = self._olecp,self._olecp_cookie,None,None
            cp.Unadvise(cookie)
    def _query_interface_(self, iid):
        import win32com.server.util
        if iid==self.CLSID_Sink: return win32com.server.util.wrap(self)

    # Event Handlers
    # If you create handlers, they should have the following prototypes:
#   def OnDefComplete(self, DefID=defaultNamedNotOptArg):


from win32com.client import CoClassBaseClass
# This CoClass is known by the name 'RPCO.X'
class RPcoX(CoClassBaseClass): # A CoClass
    # RPcoX Control
    CLSID = IID('{D323A625-1D13-11D4-8858-444553540000}')
    coclass_sources = [
        _DRPcoXEvents,
    ]
    default_source = _DRPcoXEvents
    coclass_interfaces = [
        _DRPcoX,
    ]
    default_interface = _DRPcoX

RecordMap = {
}

CLSIDToClassMap = {
    '{D323A623-1D13-11D4-8858-444553540000}' : _DRPcoX,
    '{D323A624-1D13-11D4-8858-444553540000}' : _DRPcoXEvents,
    '{D323A625-1D13-11D4-8858-444553540000}' : RPcoX,
}
CLSIDToPackageMap = {}
win32com.client.CLSIDToClass.RegisterCLSIDsFromDict( CLSIDToClassMap )
VTablesToPackageMap = {}
VTablesToClassMap = {
}


NamesToIIDMap = {
    '_DRPcoX' : '{D323A623-1D13-11D4-8858-444553540000}',
    '_DRPcoXEvents' : '{D323A624-1D13-11D4-8858-444553540000}',
}





from comtypes import client
from numpy import random
from array import array
from win32com import client as wclient
from timeit import timeit

from .actxobjects import RPcoX

com_iface = client.CreateObject('RPco.X')
win_iface = wclient.Dispatch('RPco.X')
mod_iface = RPcoX()

data = random.normal(size=100e3)
setup = "from __main__ import iface, data, array"
n = 4


def test_iface(iface):
    iface.ConnectRZ6('GB', 1)
    iface.LoadCOF('components/test_physiology_RZ6.rcx')
    iface.Run()

    print("checking to see which datatypes are supported")
    try:
        print("ndarray\t", iface.WriteTagV('speaker', 0, data))
    except:
        print("error")
    print("list\t", iface.WriteTagV('speaker', 0, data.tolist()))
    print("array\t", iface.WriteTagV('speaker', 0, array('d', data)))

    print("Testing speed of calling WriteTagV")
    try:
        t = timeit("iface.WriteTagV('speaker', 0, data)", setup, number=n)/n
        print("ndarray\t", t)
    except:
        print("error")
    t = timeit("iface.WriteTagV('speaker', 0, data.tolist())",
               setup, number=n)/n
    print("list\t", t)
    t = timeit("iface.WriteTagV('speaker', 0, array('d', data))",
               setup, number=n)/n
    print("array\t", t)

    print("Testing speed of calling ReadTagV")
    t = timeit("iface.ReadTagV('speaker', 0, 200e3)", setup, number=n)/n
    print(200e3*4/t/2**20, "MB/sec")

if __name__ == '__main__':
    # print "Testing win32com"
    # iface = win_iface
    # test_iface(iface)
    # print "Testing comtypes"
    # iface = com_iface
    # test_iface(iface)
    print("Testing modified win32com")
    iface = mod_iface
    test_iface(iface)



# -*- coding: utf-8 -*-
# Created by makepy.py version 0.5.00
# By python version 2.5.4 (r254:67916, Dec 23 2008, 15:10:54) [MSC v.1310 32 bit (Intel)]
# From type library 'PA5x.ocx'
# On Thu Aug 27 14:11:40 2009
"""PA5x ActiveX Control module"""
makepy_version = '0.5.00'
python_version = 0x20504f0

import win32com.client.CLSIDToClass, pythoncom, pywintypes
import win32com.client.util
from pywintypes import IID
from win32com.client import Dispatch

# The following 3 lines may need tweaking for the particular server
# Candidates are pythoncom.Missing, .Empty and .ArgNotFound
defaultNamedOptArg=pythoncom.Empty
defaultNamedNotOptArg=pythoncom.Empty
defaultUnnamedArg=pythoncom.Empty

CLSID = IID('{73B7CFAB-7D2C-487A-81EC-E6A15FB9E84A}')
MajorVersion = 1
MinorVersion = 0
LibraryFlags = 10
LCID = 0x0

from win32com.client import DispatchBaseClass
class _DPA5x(DispatchBaseClass):
	"""Dispatch interface for PA5x Control"""
	CLSID = IID('{9792C25C-198C-4AC3-B915-246CB15BB30C}')
	coclass_clsid = IID('{EC05FCDE-300D-4CE3-9774-A4C377507BD7}')

	def Connect(self, Interface=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(1, LCID, 1, (11, 0), ((3, 0), (3, 0)),Interface
			, DevNum)

	def ConnectPA5(self, IntName=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(8, LCID, 1, (3, 0), ((8, 0), (3, 0)),IntName
			, DevNum)

	def Display(self, Text=defaultNamedNotOptArg, Position=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(7, LCID, 1, (11, 0), ((8, 0), (3, 0)),Text
			, Position)

	def GetAtten(self):
		return self._oleobj_.InvokeTypes(3, LCID, 1, (4, 0), (),)

	def GetError(self):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(6, LCID, 1, (8, 0), (),)

	def Reset(self):
		return self._oleobj_.InvokeTypes(4, LCID, 1, (11, 0), (),)

	def SetAtten(self, AttVal=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(2, LCID, 1, (11, 0), ((4, 0),),AttVal
			)

	def SetUser(self, ParCode=defaultNamedNotOptArg, Val=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(5, LCID, 1, (11, 0), ((3, 0), (4, 0)),ParCode
			, Val)

	_prop_map_get_ = {
	}
	_prop_map_put_ = {
	}

class _DPA5xEvents:
	"""Event interface for PA5x Control"""
	CLSID = CLSID_Sink = IID('{06D79585-4000-4243-BD17-499FA2F46DDB}')
	coclass_clsid = IID('{EC05FCDE-300D-4CE3-9774-A4C377507BD7}')
	_public_methods_ = [] # For COM Server support
	_dispid_to_func_ = {
		}

	def __init__(self, oobj = None):
		if oobj is None:
			self._olecp = None
		else:
			import win32com.server.util
			from win32com.server.policy import EventHandlerPolicy
			cpc=oobj._oleobj_.QueryInterface(pythoncom.IID_IConnectionPointContainer)
			cp=cpc.FindConnectionPoint(self.CLSID_Sink)
			cookie=cp.Advise(win32com.server.util.wrap(self, usePolicy=EventHandlerPolicy))
			self._olecp,self._olecp_cookie = cp,cookie
	def __del__(self):
		try:
			self.close()
		except pythoncom.com_error:
			pass
	def close(self):
		if self._olecp is not None:
			cp,cookie,self._olecp,self._olecp_cookie = self._olecp,self._olecp_cookie,None,None
			cp.Unadvise(cookie)
	def _query_interface_(self, iid):
		import win32com.server.util
		if iid==self.CLSID_Sink: return win32com.server.util.wrap(self)

	# Event Handlers
	# If you create handlers, they should have the following prototypes:


from win32com.client import CoClassBaseClass
# This CoClass is known by the name 'PA5.X'
class PA5x(CoClassBaseClass): # A CoClass
	# PA5x Control
	CLSID = IID('{EC05FCDE-300D-4CE3-9774-A4C377507BD7}')
	coclass_sources = [
		_DPA5xEvents,
	]
	default_source = _DPA5xEvents
	coclass_interfaces = [
		_DPA5x,
	]
	default_interface = _DPA5x

RecordMap = {
}

CLSIDToClassMap = {
	'{EC05FCDE-300D-4CE3-9774-A4C377507BD7}' : PA5x,
	'{9792C25C-198C-4AC3-B915-246CB15BB30C}' : _DPA5x,
	'{06D79585-4000-4243-BD17-499FA2F46DDB}' : _DPA5xEvents,
}
CLSIDToPackageMap = {}
win32com.client.CLSIDToClass.RegisterCLSIDsFromDict( CLSIDToClassMap )
VTablesToPackageMap = {}
VTablesToClassMap = {
}


NamesToIIDMap = {
	'_DPA5xEvents' : '{06D79585-4000-4243-BD17-499FA2F46DDB}',
	'_DPA5x' : '{9792C25C-198C-4AC3-B915-246CB15BB30C}',
}





# -*- coding: utf-8 -*-
# Created by makepy.py version 0.5.00
# By python version 2.6.6 |EPD 6.3-1 (32-bit)| (r266:84292, Sep 20 2010, 11:26:16) [MSC v.1500 32 bit (Intel)]
# From type library 'TDevAccX.ocx'
# On Mon Nov 22 10:28:02 2010
"""TDevAccX ActiveX Control module"""
makepy_version = '0.5.00'
python_version = 0x20606f0

import win32com.client.CLSIDToClass, pythoncom, pywintypes
import win32com.client.util
from pywintypes import IID
from win32com.client import Dispatch

# The following 3 lines may need tweaking for the particular server
# Candidates are pythoncom.Missing, .Empty and .ArgNotFound
defaultNamedOptArg=pythoncom.Empty
defaultNamedNotOptArg=pythoncom.Empty
defaultUnnamedArg=pythoncom.Empty

CLSID = IID('{735FF2F5-CC9A-4D67-BFC1-FD5FA6742027}')
MajorVersion = 1
MinorVersion = 0
LibraryFlags = 10
LCID = 0x0

from win32com.client import DispatchBaseClass
class _DTDevAccX(DispatchBaseClass):
	"""Dispatch interface for TDevAccX Control"""
	CLSID = IID('{F04FC131-D33B-45B7-BC22-CF58B06CA5CB}')
	coclass_clsid = IID('{09EFA19D-3AD0-49A8-8232-18D6F7512CE8}')

	def AboutBox(self):
		return self._oleobj_.InvokeTypes(-552, LCID, 1, (24, 0), (),)

	def CheckServerConnection(self):
		return self._oleobj_.InvokeTypes(2, LCID, 1, (3, 0), (),)

	def CloseConnection(self):
		return self._oleobj_.InvokeTypes(3, LCID, 1, (24, 0), (),)

	def ConnectServer(self, ServerName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(1, LCID, 1, (3, 0), ((8, 0),),ServerName
			)

	def GetDeviceName(self, Index=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(17, LCID, 1, (8, 0), ((3, 0),),Index
			)

	def GetDeviceRCO(self, DeviceName=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(19, LCID, 1, (8, 0), ((8, 0),),DeviceName
			)

	def GetDeviceSF(self, Target=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(10, LCID, 1, (4, 0), ((8, 0),),Target
			)

	def GetDeviceStatus(self, Target=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(14, LCID, 1, (3, 0), ((8, 0),),Target
			)

	def GetDeviceType(self, DeviceName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(18, LCID, 1, (3, 0), ((8, 0),),DeviceName
			)

	def GetNextTag(self, DeviceName=defaultNamedNotOptArg, ReqType=defaultNamedNotOptArg, DoFirst=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(20, LCID, 1, (8, 0), ((8, 0), (3, 0), (3, 0)),DeviceName
			, ReqType, DoFirst)

	def GetSysMode(self):
		return self._oleobj_.InvokeTypes(16, LCID, 1, (3, 0), (),)

	def GetTankName(self):
		"""method GetTankName"""
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(23, LCID, 1, (8, 0), (),)

	def GetTargetClass(self, Target=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(8, LCID, 1, (3, 0), ((8, 0),),Target
			)

	def GetTargetSize(self, Target=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(11, LCID, 1, (3, 0), ((8, 0),),Target
			)

	def GetTargetType(self, Target=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(15, LCID, 1, (3, 0), ((8, 0),),Target
			)

	def GetTargetVal(self, Target=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(12, LCID, 1, (5, 0), ((8, 0),),Target
			)

	def ReadTarget(self, Target=defaultNamedNotOptArg, nOS=defaultNamedNotOptArg, nWords=defaultNamedNotOptArg, pData=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(6, LCID, 1, (3, 0), ((8, 0), (3, 0), (3, 0), (16388, 0)),Target
			, nOS, nWords, pData)

	def ReadTargetV(self, Target=defaultNamedNotOptArg, nOS=defaultNamedNotOptArg, nWords=defaultNamedNotOptArg):
		return self._ApplyTypes_(7, 1, (12, 0), ((8, 0), (3, 0), (3, 0)), u'ReadTargetV', None,Target
			, nOS, nWords)

	def ReadTargetVEX(self, Target=defaultNamedNotOptArg, nOS=defaultNamedNotOptArg, nWords=defaultNamedNotOptArg, SrcType=defaultNamedNotOptArg
			, DstType=defaultNamedNotOptArg):
		return self._ApplyTypes_(24, 1, (12, 0), ((8, 0), (3, 0), (3, 0), (8, 0), (8, 0)), u'ReadTargetVEX', None,Target
			, nOS, nWords, SrcType, DstType)

	def SetSysMode(self, NewMode=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(21, LCID, 1, (3, 0), ((3, 0),),NewMode
			)

	def SetTankName(self, TankName=defaultNamedNotOptArg):
		"""method SetTankName"""
		return self._oleobj_.InvokeTypes(22, LCID, 1, (3, 0), ((8, 0),),TankName
			)

	def SetTargetVal(self, Target=defaultNamedNotOptArg, Val=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(13, LCID, 1, (3, 0), ((8, 0), (5, 0)),Target
			, Val)

	def WriteTarget(self, Target=defaultNamedNotOptArg, nOS=defaultNamedNotOptArg, nWords=defaultNamedNotOptArg, pData=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(4, LCID, 1, (3, 0), ((8, 0), (3, 0), (3, 0), (16388, 0)),Target
			, nOS, nWords, pData)

	def WriteTargetV(self, Target=defaultNamedNotOptArg, nOS=defaultNamedNotOptArg, vData=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(5, LCID, 1, (3, 0), ((8, 0), (3, 0), (12, 0)),Target
			, nOS, vData)

	def WriteTargetVEX(self, Target=defaultNamedNotOptArg, nOS=defaultNamedNotOptArg, DstType=defaultNamedNotOptArg, Buf=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(25, LCID, 1, (3, 0), ((8, 0), (3, 0), (8, 0), (12, 0)),Target
			, nOS, DstType, Buf)

	def ZeroTarget(self, Target=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(9, LCID, 1, (24, 0), ((8, 0),),Target
			)

	_prop_map_get_ = {
	}
	_prop_map_put_ = {
	}

class _DTDevAccXEvents:
	"""Event interface for TDevAccX Control"""
	CLSID = CLSID_Sink = IID('{64E8C018-79F7-4C72-9244-6CAE7315BA4F}')
	coclass_clsid = IID('{09EFA19D-3AD0-49A8-8232-18D6F7512CE8}')
	_public_methods_ = [] # For COM Server support
	_dispid_to_func_ = {
		}

	def __init__(self, oobj = None):
		if oobj is None:
			self._olecp = None
		else:
			import win32com.server.util
			from win32com.server.policy import EventHandlerPolicy
			cpc=oobj._oleobj_.QueryInterface(pythoncom.IID_IConnectionPointContainer)
			cp=cpc.FindConnectionPoint(self.CLSID_Sink)
			cookie=cp.Advise(win32com.server.util.wrap(self, usePolicy=EventHandlerPolicy))
			self._olecp,self._olecp_cookie = cp,cookie
	def __del__(self):
		try:
			self.close()
		except pythoncom.com_error:
			pass
	def close(self):
		if self._olecp is not None:
			cp,cookie,self._olecp,self._olecp_cookie = self._olecp,self._olecp_cookie,None,None
			cp.Unadvise(cookie)
	def _query_interface_(self, iid):
		import win32com.server.util
		if iid==self.CLSID_Sink: return win32com.server.util.wrap(self)

	# Event Handlers
	# If you create handlers, they should have the following prototypes:


from win32com.client import CoClassBaseClass
# This CoClass is known by the name 'TDevAcc.X'
class TDevAccX(CoClassBaseClass): # A CoClass
	# TDevAccX Control
	CLSID = IID('{09EFA19D-3AD0-49A8-8232-18D6F7512CE8}')
	coclass_sources = [
		_DTDevAccXEvents,
	]
	default_source = _DTDevAccXEvents
	coclass_interfaces = [
		_DTDevAccX,
	]
	default_interface = _DTDevAccX

RecordMap = {
}

CLSIDToClassMap = {
	'{F04FC131-D33B-45B7-BC22-CF58B06CA5CB}' : _DTDevAccX,
	'{64E8C018-79F7-4C72-9244-6CAE7315BA4F}' : _DTDevAccXEvents,
	'{09EFA19D-3AD0-49A8-8232-18D6F7512CE8}' : TDevAccX,
}
CLSIDToPackageMap = {}
win32com.client.CLSIDToClass.RegisterCLSIDsFromDict( CLSIDToClassMap )
VTablesToPackageMap = {}
VTablesToClassMap = {
}


NamesToIIDMap = {
	'_DTDevAccXEvents' : '{64E8C018-79F7-4C72-9244-6CAE7315BA4F}',
	'_DTDevAccX' : '{F04FC131-D33B-45B7-BC22-CF58B06CA5CB}',
}





import pytest

from tdt.abstract_ring_buffer import span


def test_span():
    # Test that the number of samples span are properly calculated given the
    # cycle count and index.
    assert span(0, 0, 0, 1, 10) == 1
    assert span(0, 0, 0, 9, 10) == 9
    assert span(1, 0, 1, 1, 10) == 1
    assert span(1, 0, 1, 9, 10) == 9
    assert span(1, 5, 1, 9, 10) == 4
    assert span(1, 5, 2, 4, 10) == 9
    assert span(1, 5, 2, 5, 10) == 10

    with pytest.raises(ValueError, match='Number of slots exceeds buffer size'):
        span(0, 0, 1, 9, 10)
    with pytest.raises(ValueError, match='Number of slots exceeds buffer size'):
        span(1, 0, 2, 1, 10)
    with pytest.raises(ValueError, match='Start sample higher than end sample'):
        span(3, 0, 2, 1, 10)



import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

from .dsp_project import DSPProject
from .dsp_process import DSPProcess
from .dsp_circuit import DSPCircuit
from .dsp_buffer import DSPBuffer
from .dsp_error import DSPError

try:
    from importlib.metadata import version
    __version__ = version("TDTPy")
except Exception:
    # Catch a generic exception which will handle both PackageNotFoundError and
    # ImportError (attempting to catch PackageNotFoundError is a bit tricky if
    # importlib.metadata is not available).
    try:
        from .version import __version__
    except ImportError:
        # package is not installed. probably some munging of Python path going
        # on here.
        __version__ = '0.0.0'



# -*- coding: utf-8 -*-
# Created by makepy.py version 0.5.00
# By python version 2.5.4 (r254:67916, Dec 23 2008, 15:10:54) [MSC v.1310 32 bit (Intel)]
# From type library 'zBUSx.ocx'
# On Thu Aug 27 14:12:16 2009
"""zBUSx ActiveX Control module"""
makepy_version = '0.5.00'
python_version = 0x20504f0

import win32com.client.CLSIDToClass, pythoncom, pywintypes
import win32com.client.util
from pywintypes import IID
from win32com.client import Dispatch

# The following 3 lines may need tweaking for the particular server
# Candidates are pythoncom.Missing, .Empty and .ArgNotFound
defaultNamedOptArg=pythoncom.Empty
defaultNamedNotOptArg=pythoncom.Empty
defaultUnnamedArg=pythoncom.Empty

CLSID = IID('{10055D3E-3938-4652-B6A2-6A6A4184D18D}')
MajorVersion = 1
MinorVersion = 0
LibraryFlags = 10
LCID = 0x0

from win32com.client import DispatchBaseClass
class _DZBUSx(DispatchBaseClass):
	"""Dispatch interface for ZBUSx Control"""
	CLSID = IID('{9F05A891-D2B9-41AF-8C8E-3F4245261483}')
	coclass_clsid = IID('{79734A6C-8E6E-4998-B834-3E4E481232B0}')

	def AboutBox(self):
		return self._oleobj_.InvokeTypes(-552, LCID, 1, (24, 0), (),)

	def Connect(self, Interface=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(1, LCID, 1, (3, 0), ((3, 0),),Interface
			)

	def ConnectZBUS(self, IntName=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(12, LCID, 1, (3, 0), ((8, 0),),IntName
			)

	def FlushIO(self, RackNum=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(5, LCID, 1, (3, 0), ((3, 0),),RackNum
			)

	def GetDeviceAddr(self, DevType=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(2, LCID, 1, (3, 0), ((3, 0), (3, 0)),DevType
			, DevNum)

	def GetDeviceAt(self, RackNum=defaultNamedNotOptArg, PosNum=defaultNamedNotOptArg, DevID=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(11, LCID, 1, (8, 0), ((3, 0), (3, 0), (16387, 0), (16387, 0)),RackNum
			, PosNum, DevID, DevNum)

	def GetDeviceVersion(self, DevType=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(3, LCID, 1, (3, 0), ((3, 0), (3, 0)),DevType
			, DevNum)

	def GetError(self):
		# Result is a Unicode object
		return self._oleobj_.InvokeTypes(10, LCID, 1, (8, 0), (),)

	def HardwareReset(self, RackNum=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(4, LCID, 1, (3, 0), ((3, 0),),RackNum
			)

	def KillCode(self, DevType=defaultNamedNotOptArg, DevNum=defaultNamedNotOptArg, MagicCode=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(9, LCID, 1, (3, 0), ((3, 0), (3, 0), (3, 0)),DevType
			, DevNum, MagicCode)

	def zBusSync(self, RackMask=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(8, LCID, 1, (3, 0), ((3, 0),),RackMask
			)

	def zBusTrigA(self, RackNum=defaultNamedNotOptArg, zTrgMode=defaultNamedNotOptArg, Delay=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(6, LCID, 1, (3, 0), ((3, 0), (3, 0), (3, 0)),RackNum
			, zTrgMode, Delay)

	def zBusTrigB(self, RackNum=defaultNamedNotOptArg, zTrgMode=defaultNamedNotOptArg, Delay=defaultNamedNotOptArg):
		return self._oleobj_.InvokeTypes(7, LCID, 1, (3, 0), ((3, 0), (3, 0), (3, 0)),RackNum
			, zTrgMode, Delay)

	_prop_map_get_ = {
	}
	_prop_map_put_ = {
	}

class _DZBUSxEvents:
	"""Event interface for ZBUSx Control"""
	CLSID = CLSID_Sink = IID('{575833D5-0B5E-4759-B370-40FA23D409E5}')
	coclass_clsid = IID('{79734A6C-8E6E-4998-B834-3E4E481232B0}')
	_public_methods_ = [] # For COM Server support
	_dispid_to_func_ = {
		}

	def __init__(self, oobj = None):
		if oobj is None:
			self._olecp = None
		else:
			import win32com.server.util
			from win32com.server.policy import EventHandlerPolicy
			cpc=oobj._oleobj_.QueryInterface(pythoncom.IID_IConnectionPointContainer)
			cp=cpc.FindConnectionPoint(self.CLSID_Sink)
			cookie=cp.Advise(win32com.server.util.wrap(self, usePolicy=EventHandlerPolicy))
			self._olecp,self._olecp_cookie = cp,cookie
	def __del__(self):
		try:
			self.close()
		except pythoncom.com_error:
			pass
	def close(self):
		if self._olecp is not None:
			cp,cookie,self._olecp,self._olecp_cookie = self._olecp,self._olecp_cookie,None,None
			cp.Unadvise(cookie)
	def _query_interface_(self, iid):
		import win32com.server.util
		if iid==self.CLSID_Sink: return win32com.server.util.wrap(self)

	# Event Handlers
	# If you create handlers, they should have the following prototypes:


from win32com.client import CoClassBaseClass
# This CoClass is known by the name 'ZBUS.X'
class ZBUSx(CoClassBaseClass): # A CoClass
	# ZBUSx Control
	CLSID = IID('{79734A6C-8E6E-4998-B834-3E4E481232B0}')
	coclass_sources = [
		_DZBUSxEvents,
	]
	default_source = _DZBUSxEvents
	coclass_interfaces = [
		_DZBUSx,
	]
	default_interface = _DZBUSx

RecordMap = {
}

CLSIDToClassMap = {
	'{575833D5-0B5E-4759-B370-40FA23D409E5}' : _DZBUSxEvents,
	'{79734A6C-8E6E-4998-B834-3E4E481232B0}' : ZBUSx,
	'{9F05A891-D2B9-41AF-8C8E-3F4245261483}' : _DZBUSx,
}
CLSIDToPackageMap = {}
win32com.client.CLSIDToClass.RegisterCLSIDsFromDict( CLSIDToClassMap )
VTablesToPackageMap = {}
VTablesToClassMap = {
}


NamesToIIDMap = {
	'_DZBUSx' : '{9F05A891-D2B9-41AF-8C8E-3F4245261483}',
	'_DZBUSxEvents' : '{575833D5-0B5E-4759-B370-40FA23D409E5}',
}


